"""This module holds argument validators for ``cast()``.

See the API docs for :func:`@extension_func <pdcast.extension_func>` for more
details.
"""
# pylint: disable=unused-argument, anomalous-backslash-in-string
# pylint: disable=redefined-outer-name
from __future__ import annotations
import datetime
import decimal
from typing import Any, Callable, Iterable

import pandas as pd

from pdcast import types
from pdcast.detect import detect_type
from pdcast.patch.round import rule as rounding_rule
from pdcast.patch.snap import tol as snap_tol
from pdcast.resolve import resolve_type
from pdcast.util.round import Tolerance
from pdcast.util import time
from pdcast.util.type_hints import datetime_like, type_specifier
from pdcast.util.vector import as_series

from .base import cast


# TODO: include context dicts in parameter lists


# ignore this file when doing string-based object lookups in resolve_type()
IGNORE_FRAME_OBJECTS = True


#######################
####    PRIVATE    ####
#######################


valid_errors = ("raise", "coerce", "ignore")


def as_string_set(val: str | set[str]) -> set[str]:
    """Create a set of strings from a scalar or iterable."""
    # scalar string
    if isinstance(val, str):
        return {val}

    # iterable
    if hasattr(val, "__iter__"):
        if not all(isinstance(x, str) for x in val):
            raise TypeError(
                f"input must consist only of strings: {val}"
            )
        return set(val)

    # everything else
    return {str(val)}


def assert_sets_are_disjoint(set_1: set, set_2: set) -> None:
    """Raise a `ValueError` if the sets have any overlap."""
    if not set_1.isdisjoint(set_2):
        err_msg = "sets must be disjoint"

        intersection = set_1.intersection(set_2)
        if len(intersection) == 1:  # singular
            err_msg += (
                f"({repr(intersection.pop())} is present in both sets)"
            )
        else:  # plural
            err_msg += f"({intersection} are present in both sets)"

        raise ValueError(err_msg)


#########################
####    ARGUMENTS    ####
#########################


# initial defaults for all conversion arguments
defaults = {
    "tol": decimal.Decimal(1) / 10**6,
    "rounding": None,
    "unit": "ns",
    "step_size": 1,
    "since": "utc",
    "tz": None,
    "day_first": False,
    "year_first": False,
    "as_hours": False,
    "true": {"true", "t", "yes", "y", "on", "1"},
    "false": {"false", "f", "no", "n", "off", "0"},
    "ignore_case": True,
    "format": None,
    "base": 0,
    "call": None,
    "downcast": False,
    "errors": "raise"
}


@cast.argument
def series(val: Any, context: dict) -> pd.Series:
    """The series to convert.

    Parameters
    ----------
    val : Any
        An object to be wrapped as a :class:`pandas.Series <pandas.Series>`.
        If this is already a series, then it is used as-is.  If it is a
        numpy array, then it is converted to a series with the same dtype.
        Other iterables are interpreted as a ``dtype: object`` series, and
        scalars are converted into a ``dtype: object`` series with a single
        element.
    context : dict
        The current settings of each argument that was supplied to the
        conversion.  This is automatically inserted by the
        :class:`ExtensionFunc <pdcast.ExtensionFunc>` machinery, but is not
        used for this argument.

    Returns
    -------
    pandas.Series
        The input object as a series.

    Notes
    -----
    :meth:`Overloaded <pdcast.DispatchFunc.overload>` implementations of
    :func:`cast() <pdcast.cast>` should note that this argument is always
    supplied as a series, even if the original value was a scalar.
    """
    return as_series(val)


@cast.argument
def dtype(
    val: type_specifier | None,
    context: dict,
    supertype: type_specifier | None = None
) -> types.VectorType:
    """The type to convert to.

    Parameters
    ----------
    val : type_specifier | None
        A type specifier that determines the output type.  This can be in any
        format recognized by
        :func:`resolve_type() <pdcast.resolve.resolve_type>` and must not be
        composite.  If it is left as :data:`None <python:None>`, then the
        inferred type of the :func:`series <pdcast.convert.arguments.series>`
        will be used instead.
    context : dict
        The current settings of each argument that was supplied to the
        conversion.  This is automatically inserted by the
        :class:`ExtensionFunc <pdcast.ExtensionFunc>` machinery, and allows
        this argument to access the series being converted.
    supertype : type_specifier | None
        A type specifier to check the output type against.  If this is given,
        then this argument will reject any output type that is not a member of
        the given supertype's hierarchy.  This can be in any format recognized
        by :func:`resolve_type() <pdcast.resolve.resolve_type>`.

    Returns
    -------
    VectorType
        The resolved output type.

    Raises
    ------
    ValueError
        If the output type is composite or not a member of the supertype's
        hierarchy.  This can also be raised if no output type is given and the
        series is empty or contains only NAs.

    TODO: update with NullType once it is implemented

    TODO: Notes/Examples
    """
    # no dtype given
    if val is None:
        if "series" in context:
            series = context["series"]
            val = detect_type(series)
            # TODO: replace with NullType
            if val is None:  # series is empty or contains only NAs
                raise ValueError(
                    "cannot interpret empty series without an explicit "
                    "`dtype` argument"
                )

    else:
        val = resolve_type(val)

        # dtype is naked decorator
        if val.unwrap() is None:
            if "series" not in context:
                raise ValueError(
                    "cannot perform anonymous conversion without data"
                )
            encountered = list(val.decorators)
            val = encountered.pop(-1).replace(
                wrapped=detect_type(context["series"])
            )
            for decorator in reversed(encountered):
                val = decorator.replace(wrapped=val)

    # reject composite
    if isinstance(val, types.CompositeType):
        raise ValueError(
            f"`dtype` cannot be composite (received: {val})"
        )

    # reject improper subtype
    if supertype is not None and val.unwrap() is not None:
        supertype = resolve_type(supertype)
        if not supertype.contains(val.unwrap()):
            raise ValueError(
                f"`dtype` must be {supertype}-like, not {val}"
            )

    return val


@cast.argument(default=defaults["tol"])
def tol(val: str, context: dict) -> Tolerance:
    """The maximum amount of precision loss that can occur before an error
    is raised.

    Parameters
    ----------
    val : numeric
        A scalar numeric that is coercible to
        :class:`Decimal <python:decimal.Decimal>`.  In the case of complex
        values, their real and imaginary components are considered separately.
        Defaults to ``1e-6``

    Returns
    -------
    Tolerance
        A ``Tolerance`` object that consists of two
        :class:`Decimal <python:decimal.Decimal>` values, one for both the real
        and imaginary components.  This maintains the highest possible
        precision in both cases.

    Raises
    ------
    TypeError
        If ``val`` could not be coerced into a
        :class:`Decimal <python:decimal.Decimal>` representation.
    ValueError
        If the real or imaginary component of ``val`` is not positive.

    Notes
    -----
    Precision loss is defined using a 2-sided window around each of the
    observed values.  The size of this window is directly controlled by
    this argument.  If a conversion causes any value to be coerced outside
    this window, then a :class:`ValueError <python:ValueError>` will be raised.

    This argument only affects numeric conversions.

    Examples
    --------
    The input to this argument must be a positive numeric that is
    coercible to :class:`Decimal <python:decimal.Decimal>`.

    .. doctest::

        >>> cast(1.001, "int", tol=0.01)
        0    1
        dtype: int64
        >>> cast(1.001, "int", tol=0)
        Traceback (most recent call last):
            ...
        ValueError: precision loss exceeds tolerance 0 at index [0]

    If a complex value is given, then its real and imaginary components
    will be considered separately.

    .. doctest::

        >>> cast(1.001+0.001j, "int", tol=0.01+0.01j)
        0    1
        dtype: int64
        >>> cast(1.001+0.001j, "int", tol=0.01+0j)
        Traceback (most recent call last):
            ...
        ValueError: imaginary component exceeds tolerance 0 at index [0]

    This argument also has special behavior around the min/max of bounded
    numerics, like integers and booleans.  If a value would normally
    overflow, but falls within tolerance of these bounds, then it will be
    clipped to fit rather than raise an
    :class:`OverflowError <python:OverflowError>`.

    .. doctest::

        >>> cast(129, "int8", tol=2)
        0    127
        dtype: int8
        >>> cast(129, "int8", tol=0)
        Traceback (most recent call last):
            ...
        OverflowError: values exceed int8 range at index [0]

    Additionally, this argument controls the maximum amount of precision
    loss that can occur when
    :func:`downcasting <pdcast.convert.arguments.downcast>` numeric values.

    .. doctest::

        >>> cast(1.1, "float", tol=0, downcast=True)
        0    1.1
        dtype: float64
        >>> cast(1.1, "float", tol=0.001, downcast=True)
        0    1.099609
        dtype: float16

    Setting this to infinity ignores precision loss entirely.

    .. doctest::

        >>> cast(1.5, "int", tol=np.inf)
        0    2
        dtype: int64
        >>> cast(np.inf, "int64", tol=np.inf)
        0    9223372036854775807
        dtype: int64

    .. note::

        For integer conversions, this is equivalent to setting
        :func:`rounding <pdcast.convert.arguments.rounding>` to
        ``"half_even"``, with additional clipping around the minimum and
        maximum values.
    """
    return snap_tol(val, context)


@cast.argument(default=defaults["rounding"])
def rounding(val: str, context: dict) -> str:
    """The rounding rule to use for numeric conversions.

    Parameters
    ----------
    val : str | None
        An optional string specifying the rounding rule to use, or :data:`None`
        to indicate that no rounding will be applied.  Defaults to :data:`None`.

    Returns
    -------
    str | None
        A validated version of the string passed to ``val`` or :data:`None`.

    Raises
    ------
    TypeError
        If ``val`` is not a string or :data:`None <python:None>`.
    ValueError
        If ``val`` does not correspond to one of the recognized rounding rules.

    Notes
    -----
    The available options for this argument are as follows:

        *   ``None`` - do not round.
        *   ``"floor"`` - round toward negative infinity.
        *   ``"ceiling"`` - round toward positive infinity.
        *   ``"down"`` - round toward zero.
        *   ``"up"`` - round away from zero.
        *   ``"half_floor"`` - round to nearest with ties toward positive infinity.
        *   ``"half_ceiling"`` - round to nearest with ties toward negative
            infinity.
        *   ``"half_down"`` - round to nearest with ties toward zero.
        *   ``"half_up"`` - round to nearest with ties away from zero.
        *   ``"half_even"`` - round to nearest with ties toward the `nearest even
            value <https://en.wikipedia.org/wiki/Rounding#Rounding_half_to_even>`_.
            Also known as *convergent rounding*, *statistician's rounding*, or
            *banker's rounding*.

    This argument is applied **after**
    :func:`tol <pdcast.convert.arguments.tol>`.

    Examples
    --------
    .. doctest::

        >>> cast([-1.5, -0.5, 0.2, 1.7], "int", rounding="floor")
        0   -2
        1   -1
        2    0
        3    1
        dtype: int64
        >>> cast([-1.5, -0.5, 0.2, 1.7], "int", rounding="ceiling")
        0   -1
        1    0
        2    1
        3    2
        dtype: int64
        >>> cast([-1.5, -0.5, 0.2, 1.7], "int", rounding="down")
        0   -1
        1    0
        2    0
        3    1
        dtype: int64
        >>> cast([-1.5, -0.5, 0.2, 1.7], "int", rounding="up")
        0   -2
        1   -1
        2    1
        3    2
        dtype: int64
        >>> cast([-1.5, -0.5, 0.2, 1.7], "int", rounding="half_floor")
        0   -2
        1   -1
        2    0
        3    2
        dtype: int64
        >>> cast([-1.5, -0.5, 0.2, 1.7], "int", rounding="half_ceiling")
        0   -1
        1    0
        2    0
        3    2
        dtype: int64
        >>> cast([-1.5, -0.5, 0.2, 1.7], "int", rounding="half_down")
        0   -1
        1    0
        2    0
        3    2
        dtype: int64
        >>> cast([-1.5, -0.5, 0.2, 1.7], "int", rounding="half_up")
        0   -2
        1   -1
        2    0
        3    2
        dtype: int64
        >>> cast([-1.5, -0.5, 0.2, 1.7], "int", rounding="half_even")
        0   -2
        1    0
        2    0
        3    2
        dtype: int64
    """
    return rounding_rule(val, context)


@cast.argument(default=defaults["unit"])
def unit(val: str, context: dict) -> str:
    """The unit to use for numeric <-> datetime/timedelta conversions.

    Parameters
    ----------
    val : str
        A string specifying the unit to use during conversions.  Defaults to
        ``"ns"``.

    Returns
    -------
    str
        A validated version of the string passed to ``val``.

    Raises
    ------
    TypeError
        If `val` is not a string.
    ValueError
        If ``val`` does not correspond to one of the recognized units.

    See Also
    --------
    step_size <pdcast.convert.arguments.step_size> :
        The step size to use for each unit.

    Notes
    -----
    The available options for this argument are as follows (from smallest
    to largest):

        * ``"ns"`` - nanoseconds
        * ``"us"`` - microseconds
        * ``"ms"`` - milliseconds
        * ``"s"`` - seconds
        * ``"m"`` - minutes
        * ``"h"`` - hours
        * ``"D"`` - days
        * ``"W"`` - weeks
        * ``"M"`` - months
        * ``"Y"`` - years

    Examples
    --------
    .. doctest::

        >>> cast(1, "datetime", unit="ns")
        0   1970-01-01 00:00:00.000000001
        dtype: datetime64[ns]
        >>> cast(1, "datetime", unit="us")
        0   1970-01-01 00:00:00.000001
        dtype: datetime64[ns]
        >>> cast(1, "datetime", unit="ms")
        0   1970-01-01 00:00:00.001
        dtype: datetime64[ns]
        >>> cast(1, "datetime", unit="s")
        0   1970-01-01 00:00:01
        dtype: datetime64[ns]
        >>> cast(1, "datetime", unit="m")
        0   1970-01-01 00:01:00
        dtype: datetime64[ns]
        >>> cast(1, "datetime", unit="h")
        0   1970-01-01 01:00:00
        dtype: datetime64[ns]
        >>> cast(1, "datetime", unit="D")
        0   1970-01-02
        dtype: datetime64[ns]
        >>> cast(1, "datetime", unit="W")
        0   1970-01-08
        dtype: datetime64[ns]

    Units ``"M"`` and ``"Y"`` have irregular lengths.  Rather than average
    these like :func:`pandas.to_datetime`, :func:`cast() <pdcast.cast>` gives
    calendar-accurate results.

    .. doctest::

        >>> cast(1, "datetime", unit="M")
        0   1970-02-01
        dtype: datetime64[ns]
        >>> cast(1, "datetime", unit="Y")
        0   1971-01-01
        dtype: datetime64[ns]

    This accounts for leap years as well, following the `Gregorian calendar
    <https://en.wikipedia.org/wiki/Gregorian_calendar>`_.

    .. doctest::

        >>> cast(1, "datetime", unit="M", since="1972-02")
        0   1972-03-01
        dtype: datetime64[ns]
        >>> cast(1, "datetime", unit="Y", since="1972")
        0   1973-01-01
        dtype: datetime64[ns]
    """
    if not isinstance(val, str):
        raise TypeError(f"`unit` must be a string, not {repr(val)}")
    if val not in time.valid_units:
        raise ValueError(
            f"`unit` must be one of {time.valid_units}, not {repr(val)}"
        )
    return val


@cast.argument(default=defaults["step_size"])
def step_size(val: int, context: dict) -> int:
    """The step size to use for each
    :func:`unit <pdcast.convert.arguments.unit>`.

    Parameters
    ----------
    val : int
        A positive integer >= 1.  This is effectively a multiplier for
        :func:`unit <pdcast.convert.arguments.unit>`.  Defaults to ``1``.

    Returns
    -------
    int
        A validated version of the integer passed to ``val``. 

    Raises
    ------
    TypeError
        If ``val`` is not an integer.
    ValueError
        If ``val`` is not >= 1.

    See Also
    --------
    unit <pdcast.convert.arguments.unit> :
        The unit to use for the conversion.

    Examples
    --------
    .. doctest::

        >>> cast(1, "datetime", unit="ns", step_size=5)
        0   1970-01-01 00:00:00.000000005
        dtype: datetime64[ns]
        >>> cast(1, "datetime", unit="s", step_size=30)
        0   1970-01-01 00:00:30
        dtype: datetime64[ns]
        >>> cast(1, "datetime", unit="M", step_size=3)
        0   1970-04-01
        dtype: datetime64[ns]
    """
    if not isinstance(val, int):
        raise TypeError(f"`step_size` must be an integer, not {val}")
    if val < 1:
        raise ValueError(f"`step_size` must be >= 1, not {val}")
    return val


@cast.argument(default=defaults["since"])
def since(val: str | datetime_like | time.Epoch, context: dict) -> time.Epoch:
    """The epoch to use for datetime/timedelta conversions.

    Parameters
    ----------
    val : str | datetime-like
        A datetime string, datetime object, or special epoch specifier as shown
        below.  Defaults to ``"utc"``.

    Returns
    -------
    Epoch
        An :class:`Epoch` object that represents a nanosecond offset from the
        UTC epoch (1970-01-01 00:00:00).

    Raises
    ------
    TypeError
        If ``val`` could not be interpreted as an :class:`Epoch`.

    Notes
    -----
    This can accept any datetime-like value, including datetime strings,
    actual datetime objects, or one of the special values listed below
    (from earliest to latest):

        *   ``"julian"``: refers to the start of the `Julian
            <https://en.wikipedia.org/wiki/Julian_calendar>`_ period, which
            corresponds to a historical date of January 1st, 4713 BC
            (according to the `proleptic Julian calendar
            <https://en.wikipedia.org/wiki/Proleptic_Julian_calendar>`_) or
            November 24, 4714 BC (according to the `proleptic Gregorian
            calendar <https://en.wikipedia.org//wiki/Proleptic_Gregorian_calendar>`_)
            at noon.
        *   ``"gregorian"``: refers to October 14th, 1582, the date when
            Pope Gregory XIII first instituted the `Gregorian calendar
            <https://en.wikipedia.org/wiki/Gregorian_calendar>`_.
        *   ``"ntfs"``: refers to January 1st, 1601.  Used by Microsoft's
            `NTFS <https://en.wikipedia.org/wiki/NTFS>`_ file management
            system.
        *   ``"modified julian"``: equivalent to ``"reduced julian"``
            except that it increments at midnight rather than noon.  This
            was originally introduced by the `Smithsonian Astrophysical
            Observatory <https://en.wikipedia.org/wiki/Smithsonian_Astrophysical_Observatory>`_
            to track the orbit of Sputnik, the first man-made satellite to
            orbit Earth.
        *   ``"reduced julian"``: refers to noon on November 16th, 1858,
            which drops the first two leading digits of the corresponding
            ``"julian"`` day number.
        *   ``"lotus"``: refers to December 31st, 1899, which was
            incorrectly identified as January 0, 1900 in the original
            `Lotus 1-2-3 <https://en.wikipedia.org/wiki/Lotus_1-2-3>`_
            implementation.  Still used internally in a variety of
            spreadsheet applications, including Microsoft Excel, Google
            Sheets, and LibreOffice.
        *   ``"ntp"``: refers to January 1st, 1900, which is used for
            `Network Time Protocol (NTP)
            <https://en.wikipedia.org/wiki/Network_Time_Protocol>`_
            synchronization, `IBM CICS
            <https://en.wikipedia.org/wiki/CICS>`_, `Mathematica
            <https://en.wikipedia.org/wiki/Wolfram_Mathematica>`_,
            `Common Lisp <https://en.wikipedia.org/wiki/Common_Lisp>`_, and
            the `RISC <https://en.wikipedia.org/wiki/RISC_OS>`_ operating
            system.  Can also be referred to through the ``"risc"`` alias.
        *   ``"labview"``: refers to January 1st, 1904, which is used by
            the `LabVIEW <https://en.wikipedia.org/wiki/LabVIEW>`_
            laboratory control software.
        *   ``"sas"``: refers to January 1st, 1960, which is used by the
            `SAS <https://en.wikipedia.org/wiki/SAS_(software)>`_
            statistical analysis suite.
        *   ``"utc"``: refers to January 1st, 1970, the universal
            `Unix <https://en.wikipedia.org/wiki/Unix>`_\ /\ `POSIX
            <https://en.wikipedia.org/wiki/POSIX>`_ epoch.  Can also be
            referred to through the ``"unix"`` and ``"posix"`` aliases.
        *   ``"fat"``: refers to January 1st, 1980, which is used by the
            `FAT <https://en.wikipedia.org/wiki/File_Allocation_Table>`_
            file management system, as well as `ZIP
            <https://en.wikipedia.org/wiki/ZIP_(file_format)>`_ and its
            derivatives.  Equivalent to ``"zip"``.
        *   ``"gps"``: refers to January 6th, 1980, which is used in most
            `GPS <https://en.wikipedia.org/wiki/Global_Positioning_System>`_
            systems.
        *   ``"j2000"``: refers to noon on January 1st, 2000, which is
            commonly used in astronomical applications, as well as in
            `PostgreSQL <https://en.wikipedia.org/wiki/PostgreSQL>`_.
        *   ``"cocoa"``: refers to January 1st, 2001, which is used in
            Apple's `Cocoa <https://en.wikipedia.org/wiki/Cocoa_(API)>`_
            framework for macOS and related mobile devices.
        *   ``"now"``: refers to the current `system time
            <https://en.wikipedia.org/wiki/System_time>`_ at the time
            :class:`cast` was invoked.

    .. note::

        By convention, ``"julian"``, ``"reduced_julian"``, and ``"j2000"``
        dates increment at noon (12:00:00 UTC) on the corresponding day.

    Examples
    --------
    Using epoch aliases:

    .. doctest::

        >>> cast(1, "datetime", unit="s", since="j2000")
        0   2000-01-01 12:00:01
        dtype: datetime64[ns]
        >>> cast(1, "datetime", unit="s", since="gregorian")
        0    1582-10-14 00:00:01
        dtype: datetime[python]
        >>> cast(1, "datetime", unit="s", since="julian")
        0    -4713-11-24T12:00:01.000000
        dtype: object

    Using datetime strings:

    .. doctest::

        >>> cast(1, "datetime", unit="s", since="2022-03-27")
        0   2022-03-27 00:00:01
        dtype: datetime64[ns]
        >>> cast(1, "datetime", unit="s", since="27 mar 2022")
        0   2022-03-27 00:00:01
        dtype: datetime64[ns]
        >>> cast(1, "datetime", unit="s", since="03/27/22")
        0   2022-03-27 00:00:01
        dtype: datetime64[ns]

    Using datetime objects:

    .. doctest::

        >>> cast(1, "datetime", unit="s", since=pd.Timestamp("2022-03-27"))
        0   2022-03-27 00:00:01
        dtype: datetime64[ns]
        >>> cast(1, "datetime", unit="s", since=datetime.datetime(2022, 3, 27))
        0   2022-03-27 00:00:01
        dtype: datetime64[ns]
        >>> cast(1, "datetime", unit="s", since=np.datetime64("2022-03-27"))
        0   2022-03-27 00:00:01
        dtype: datetime64[ns]
    """
    if isinstance(val, time.Epoch):
        return val

    if isinstance(val, str) and val not in time.epoch_aliases:
        val = cast(val, "datetime")
        if len(val) != 1:
            raise ValueError("`since` must be scalar")
        val = val[0]

    try:
        return time.Epoch(val)
    except Exception as err:
        raise TypeError(f"`since` must be datetime-like: {val}") from err


@cast.argument(default=defaults["tz"])
def tz(
    val: str | datetime.tzinfo | None,
    context: dict
) -> datetime.tzinfo:
    """Specifies a time zone to use for datetime conversions.

    Parameters
    ----------
    val : str | pytz.timezone | None
        An IANA time zone string, a `pytz <https://pypi.org/project/pytz/>`_
        timezone object, or :data:`None <python:None>` to indicate naive
        output.  This can also be the special string ``"local"``, which refers
        to the local time zone of the current system at the time of execution.
        Defaults to :data:`None <python:None>`.

    Returns
    -------
    pytz.timezone | None
        A `pytz <https://pypi.org/project/pytz/>`_ timezone object
        corresponding to the input.  :data:`None` indicates naive output.
        Defaults to :data:`None`.

    Raises
    ------
    pytz.exceptions.UnknownTimeZoneError
        If ``val`` could not be recognized as a time zone specifier.

    Examples
    --------
    Time zone localization is a somewhat complicated process, with
    different behavior depending on the input data type.

    Numerics (:doc:`boolean </content/types/boolean>`,
    :doc:`integer </content/types/integer>`,
    :doc:`float </content/types/float>`,
    :doc:`complex </content/types/complex>`, and
    :doc:`decimal </content/types/decimal>`) as well as
    :doc:`timedeltas </content/types/timedelta>` are always computed in UTC
    relative to the :func:`since <pdcast.convert.arguments.since>` argument.
    When a time zone is supplied via :func:`tz <pdcast.convert.arguments.tz>`,
    the resulting datetimes will be *converted* from UTC to the specified time
    zone.

    .. doctest::

        >>> cast(0, "datetime", tz="US/Pacific")
        0   1969-12-31 16:00:00-08:00
        dtype: datetime64[ns, US/Pacific]
        >>> cast(0, "datetime", since="2022-03-27", tz="Asia/Hong_Kong")
        0   2022-03-27 08:00:00+08:00
        dtype: datetime64[ns, Asia/Hong_Kong]
        >>> cast(0, "datetime", since="2022-03-27 00:00:00+0800", tz="Asia/Hong_Kong")
        0   2022-03-27 00:00:00+08:00
        dtype: datetime64[ns, Asia/Hong_Kong]

    :doc:`Strings </content/types/string>` and
    :doc:`datetimes </content/types/datetime>`, on the other hand, will be
    *localized* directly to the final :func:`tz <pdcast.convert.arguments.tz>`
    if they are naive, or *converted* to it if they are timezone-aware.

    .. doctest::

        >>> cast(pd.Timestamp(0), "datetime", tz="US/Pacific")
        0   1970-01-01 00:00:00-08:00
        dtype: datetime64[ns, US/Pacific]
        >>> cast(pd.Timestamp(0, tz="UTC"), "datetime", tz="US/Pacific")
        0   1969-12-31 16:00:00-08:00
        dtype: datetime64[ns, US/Pacific]
        >>> cast("2022-03-27", "datetime", tz="Asia/Hong_Kong")
        0   2022-03-27 00:00:00+08:00
        dtype: datetime64[ns, Asia/Hong_Kong]
        >>> cast("2022-03-27 00:00:00+0000", "datetime", tz="Asia/Hong_Kong")
        0   2022-03-27 08:00:00+08:00
        dtype: datetime64[ns, Asia/Hong_Kong]

    The same behavior as for numerics can be obtained by setting
    :func:`tz <pdcast.convert.arguments.tz>` to ``"utc"`` and adding an extra
    :meth:`tz_convert() <pandas.Series.dt.tz_convert>` step, as follows:

    .. doctest::

        >>> cast(pd.Timestamp(0), "datetime", tz="utc").dt.tz_convert("US/Pacific")
        0   1969-12-31 16:00:00-08:00
        dtype: datetime64[ns, US/Pacific]
        >>> cast("2022-03-27", "datetime", tz="utc").dt.tz_convert("Asia/Hong_Kong")
        0   2022-03-27 08:00:00+08:00
        dtype: datetime64[ns, Asia/Hong_Kong]

    .. note::

        In this sense, :func:`tz <pdcast.convert.arguments.tz>` is similar to
        the ``utc`` argument of :func:`pandas.to_datetime`, but allows for full
        control over the handling of naive inputs.
    """
    return time.tz(val, context)


@cast.argument(default=defaults["day_first"])
def day_first(val: bool, context: dict) -> bool:
    """Indicates whether to interpret the first value in an ambiguous
    3-integer date (e.g. 01/05/09) as the day (``True``) or month
    (``False``).

    Parameters
    ----------
    val : bool
        A boolean (or boolean-like) value indicating the rule to apply for
        ambiguous string dates.  Defaults to ``False``.

    Returns
    -------
    bool
        The boolean equivalent of the input.

    See Also
    --------
    year_first <pdcast.convert.arguments.year_first> :
        The year complement of this argument.

    Notes
    -----
    This argument is equivalent to
    `dateutil's <https://dateutil.readthedocs.io/en/stable>`_
    :class:`dayfirst <dateutil.parser.parserinfo>` argument.

    Examples
    --------
    By default, dateutil parses ambiguous strings in MDY (American) format.

    .. doctest::

        >>> cast("01/05/09", "datetime")
        0   2009-01-05
        dtype: datetime64[ns]

    Setting this argument to ``True`` changes this to DMY (international)
    format.

    .. doctest::

        >>> cast("01/05/09", "datetime", day_first=True)
        0   2009-05-01
        dtype: datetime64[ns]

    If :func:`year_first <pdcast.convert.arguments.year_first>` is set to
    ``True``, then this argument distinguishes between YMD and YDM.

    .. doctest::

        >>> cast("01/05/09", "datetime", day_first=False, year_first=True)
        0   2001-05-09
        dtype: datetime64[ns]
        >>> cast("01/05/09", "datetime", day_first=True, year_first=True)
        0   2001-09-05
        dtype: datetime64[ns]
    """
    return bool(val)


@cast.argument(default=defaults["year_first"])
def year_first(val: bool, context: dict) -> bool:
    """Indicates whether to interpret the first value in an ambiguous
    3-integer date (e.g. 01/05/09) as the year.

    Parameters
    ----------
    val : bool
        A boolean (or boolean-like) value indicating the rule to apply for
        ambiguous string dates.  Defaults to ``False``.

    Returns
    -------
    bool
        The boolean equivalent of the input.

    See Also
    --------
    day_first <pdcast.convert.arguments.day_first> :
        The day complement of this argument.

    Notes
    -----
    This argument is equivalent to
    `dateutil's <https://dateutil.readthedocs.io/en/stable>`_
    :class:`yearfirst <dateutil.parser.parserinfo>` argument.

    Examples
    --------
    By default, dateutil interprets the last number in an ambiguous date string
    to be the year.

    .. doctest::

        >>> cast("01/05/09", "datetime")
        0   2009-01-05
        dtype: datetime64[ns]

    Setting this to ``True`` takes the first number as the year.

    .. doctest::

        >>> cast("01/05/09", "datetime", year_first=True)
        0   2001-05-09
        dtype: datetime64[ns]

    See the :func:`day_first <pdcast.convert.arguments.day_first>` argument
    for examples on how these arguments interact.
    """
    return bool(val)


@cast.argument(default=defaults["as_hours"])
def as_hours(val: bool, context: dict) -> bool:
    """Indicates whether to interpret ambiguous MM:SS timedeltas as HH:MM.

    Parameters
    ----------
    val : bool
        A boolean (or boolean-like) value indicating the rule to apply for
        ambiguous string timedeltas.  Defaults to ``False``.

    Returns
    -------
    bool
        The boolean equivalent of the input.

    Examples
    --------
    :func:`cast() <pdcast.cast>` supports a variety of timedelta string formats,
    including those given as HH:MM:SS.  Oftentimes, these strings are provided
    without a third component, which causes them to become ambiguous.  By
    default, :func:`cast() <pdcast.cast>` interprets them as minutes and seconds.

    .. doctest::

        >>> cast("1:22", "timedelta")
        0   0 days 00:01:22
        dtype: timedelta64[ns]

    Setting this argument to ``True`` changes this assumption, taking them
    to be hours and minutes instead.

    .. doctest::

        >>> cast("1:22", "timedelta", as_hours=True)
        0   0 days 01:22:00
        dtype: timedelta64[ns]
    """
    return bool(val)


@cast.argument(default=defaults["true"])
def true(val: str | Iterable[str] | None, context: dict) -> set[str]:
    """A set of truthy strings to use for boolean conversions.

    Parameters
    ----------
    val : str | Iterable[str] | None
        A string, sequence of strings, or :data:`None <python:None>`, to be
        converted into a set.  :data:`None` indicates an empty set, and scalar
        strings are converted into sets of length 1.  This can also include the
        special character ``"*"`` as a catch-all wildcard.  Defaults to
        ``{"true", "t", "yes", "y", "on", "1"}``.
    context : dict
        A dictionary containing the values of other arguments, for comparison
        with :func:`false <pdcast.convert.arguments.false>`.

    Returns
    -------
    set[str]
        A set of strings to consider ``True`` during boolean conversions. The
        returned set must be disjoint with
        :func:`false <pdcast.convert.arguments.false>`, and any strings that
        are not contained in either set will raise an error.

    Raises
    ------
    TypeError
        If ``val`` does not contain strings.
    ValueError
        If one or more elements of the set intersect with
        :func:`false <pdcast.convert.arguments.false>`.

    See Also
    --------
    false <pdcast.convert.arguments.false> :
        the falsy equivalent of this argument.
    ignore_case <pdcast.convert.arguments.ignore_case> :
        whether to ignore differences of case during boolean comparisons.

    Examples
    --------
    Converting between strings and booleans is a bit unintuitive thanks to
    Python's rules around :ref:`truth value testing <python:truth>`.  Since
    strings are counted as sequences, only empty strings will ever explicitly
    evaluate to ``False``.

    .. doctest::

        >>> bool("")
        False
        >>> bool("False")
        True

    As a result, we have to implement a more complicated string parsing
    algorithm to recognize boolean values.  In ``pdcast``, this is accomplished
    through the :func:`true <pdcast.convert.arguments.true>` and
    :func:`false <pdcast.convert.arguments.false>` arguments, which define the
    strings to consider as ``True`` and ``False``, respectively.

    By default, :func:`cast() <pdcast.cast>` considers the following strings to
    be boolean-like (:func:`ignoring <pdcast.convert.arguments.ignore_case>`
    case):

    .. doctest::

        >>> cast(["true", "t", "yes", "y", "on", "1"], "bool")
        0    True
        1    True
        2    True
        3    True
        4    True
        5    True
        dtype: bool
        >>> cast(["false", "f", "no", "n", "off", "0"], "bool")
        0    False
        1    False
        2    False
        3    False
        4    False
        5    False
        dtype: bool

    Any string that falls outside these sets will raise a
    :class:`ValueError <python:ValueError>`.

    .. doctest::

        >>> cast("abc", "bool")
        Traceback (most recent call last):
            ...
        ValueError: encountered non-boolean value: 'abc'

    We can change this behavior by adding the offending string to one of
    either :func:`true <pdcast.convert.arguments.true>` or
    :func:`false <pdcast.convert.arguments.false>`.

    .. doctest::

        >>> cast("abc", "bool", true=pdcast.cast.true | {"abc"})
        0    True
        dtype: bool

    Or we can coerce it into a missing value by setting
    :func:`errors <pdcast.convert.arguments.errors>` to ``"coerce"``.

    .. doctest::

        >>> cast("abc", "bool", errors="coerce")
        0    <NA>
        dtype: boolean

    Additionally, both sets support the special value ``"*"``, which acts as a
    wildcard, matching any string that is not found in either set.

    .. doctest::

        >>> cast("abc", "bool", false="*")
        0    False
        dtype: bool

    This can be used together with an explicit empty string to replicate the
    behavior of Python's :class:`bool() <python:bool>` function.

    .. doctest::

        >>> cast(["False", "", "abc"], "bool", true="*", false="")
        0     True
        1    False
        2     True
        dtype: bool
    """
    # convert to string sets
    true_set = set() if val is None else as_string_set(val)
    false_set = as_string_set(context.get("false", set()))

    # ensure sets are disjoint
    try:
        assert_sets_are_disjoint(true_set, false_set)
    except ValueError as err:
        raise ValueError("`true` and `false` must be disjoint") from err

    # apply ignore_case
    if bool(context.get("ignore_case", True)):
        true_set = {x.lower() for x in true_set}
    return true_set


@cast.argument(default=defaults["false"])
def false(val, context: dict) -> set[str]:
    """A set of falsy strings to use for boolean conversions.

    Parameters
    ----------
    val : str | Iterable[str] | None
        A string, sequence of strings, or :data:`None <python:None>`, to be
        converted into a set.  :data:`None` indicates an empty set, and scalar
        strings are converted into sets of length 1.  This can also include the
        special character ``"*"`` as a catch-all wildcard.  Defaults to
        ``{"false", "f", "no", "n", "off", "0"}``.
    context : dict
        A dictionary containing the values of other arguments, for comparison
        with :func:`true <pdcast.convert.arguments.true>`.

    Returns
    -------
    set[str]
        A set of strings to consider ``False`` during boolean conversions. The
        returned set must be disjoint with
        :func:`true <pdcast.convert.arguments.true>`, and any strings that
        are not contained in either set will raise an error.

    Raises
    ------
    TypeError
        If ``val`` does not contain strings.
    ValueError
        If one or more elements of the set intersect with
        :func:`true <pdcast.convert.arguments.true>`.

    See Also
    --------
    true <pdcast.convert.arguments.true> :
        the truthy equivalent of this argument.
    ignore_case <pdcast.convert.arguments.ignore_case> :
        whether to ignore differences of case during boolean comparisons.

    Examples
    --------
    See the docs for the :func:`true <pdcast.convert.arguments.true>` argument
    for examples on how to customize boolean conversions using these arguments.
    """
    # convert to string sets
    true_set = as_string_set(context.get("true", set()))
    false_set = set() if val is None else as_string_set(val)

    # ensure sets are disjoint
    try:
        assert_sets_are_disjoint(true_set, false_set)
    except ValueError as err:
        raise ValueError("`true` and `false` must be disjoint") from err

    # apply ignore_case
    if bool(context.get("ignore_case", True)):
        false_set = {x.lower() for x in false_set}
    return false_set


@cast.argument(default=defaults["ignore_case"])
def ignore_case(val: bool, context: dict) -> bool:
    """Indicates whether to ignore differences in case during string
    conversions.

    Parameters
    ----------
    val : bool
        A boolean (or boolean-like) value indicating the rule to apply for
        string case comparisons.  Defaults to ``True``.

    Returns
    -------
    bool
        The boolean equivalent of the input.

    See Also
    --------
    :func:`true <pdcast.convert.arguments.true>` :
        uses this argument to compare for boolean ``True`` strings.
    :func:`false <pdcast.convert.arguments.false>` :
        uses this argument to compare for boolean ``False`` strings.

    Examples
    --------
    By default, this only applies to comparisons against the sets provided in
    the :func:`true <pdcast.convert.arguments.true>` and
    :func:`true <pdcast.convert.arguments.true>` arguments, but it can
    also be intercepted by other conversions in
    :ref:`extensions <tutorial.conversions>` to the ``pdcast`` type system.

    .. doctest::

        >>> cast("True", "bool", ignore_case=True)
        0    True
        dtype: bool
        >>> cast("True", "bool", ignore_case=False)
        Traceback (most recent call last):
            ...
        ValueError: encountered non-boolean value: 'True'
    """
    return bool(val)


@cast.argument(default=defaults["format"])
def format(val: str | None, context: dict) -> str:
    """A :ref:`format specifier <python:formatspec>` to use for string
    conversions.

    Parameters
    ----------
    val : str | None
        An optional :ref:`format specification <python:formatspec>` string,
        similar to those used in `f-string <https://peps.python.org/pep-0498/>`_
        or :ref:`strptime <python:strftime-strptime-behavior>` formatting.
        Defaults to :data:`None <python:None>`.

    Returns
    -------
    str | None
        A validated version of the string passed to ``val``.

    Raises
    ------
    TypeError
        If the passed value is not a string.

    See Also
    --------
    base <pdcast.convert.arguments.base> :
        mathematical bases for string <-> integer conversions.

    Examples
    --------
    This argument can be used to provide `f-string
    <https://peps.python.org/pep-0498/>`_ format codes to conversions, for
    instance by aligning/padding results, adding commas for large numbers,
    or changing the representation of fractional numerics.

    .. doctest::

        >>> cast("Hello, World!", "string", format="_^24")
        0    _____Hello, World!______
        dtype: string
        >>> cast(1000000, "string", format=",")
        0    1,000,000
        dtype: string
        >>> cast(1 / 3, "string", format=".2e")
        0    3.33e-01
        dtype: string

    It can also be used to provide explicit datetime formats for
    :meth:`strptime() <python:datetime.datetime.strptime>`.

    .. doctest::

        >>> cast("7:00 AM 01/05/09", "datetime", format="%I:%M %p %d/%m/%y")
        0   2009-05-01 07:00:00
        dtype: datetime64[ns]
    """
    if val is not None and not isinstance(val, str):
        raise TypeError(f"`format` must be a string, not {val}")
    return val


@cast.argument(default=defaults["base"])
def base(val: int, context: dict) -> int:
    """Base to use for integer <-> string conversions, as supplied to
    :class:`int() <python:int>`.

    Parameters
    ----------
    val : int
        An integer representing the base to use when translating between
        string and integer formats.  This must be either 0 or between 2 and
        36, equal to the allowable bases for the :class:`int() <python:int>`
        function.  Defaults to ``0``.

    Returns
    -------
    int
        The validated equivalent of ``val``.

    Raises
    ------
    TypeError
        If ``val`` is not an integer.
    ValueError
        If ``val`` is not 0 or between 2 and 36 (inclusive).

    See Also
    --------
    format <pdcast.convert.arguments.format> :
        :ref:`format specification <python:formatspec>` for string conversions.

    Notes
    -----
    For base 0, strings are interpreted in a way similar to an
    :ref:`integer literal <python:integers>` in code, in that the actual base
    is 2, 8, 10, or 16 as determined by the prefix.  Base 0 also disallows
    leading zeros: ``int('010', 0)`` is not legal, while ``int('010')`` and
    ``int('010', 8)`` are.

    Examples
    --------
    By default, this argument will automatically parse binary, octal, or
    hexadecimal strings with the appropriate prefix.

    .. doctest::

        >>> cast("0b101", "int")
        0    5
        dtype: int64
        >>> cast("0o77", "int")
        0    63
        dtype: int64
        >>> cast("0x3FF", "int")
        0    1023
        dtype: int64

    It can also be explicitly set to deal with strings that have no prefix,
    or are in a non-standard base.

    .. doctest::

        >>> cast("101", "int", base=2)
        0    5
        dtype: int64
        >>> cast("77", "int", base=8)
        0    63
        dtype: int64
        >>> cast("3FF", "int", base=16)
        0    1023
        dtype: int64
        >>> cast("abc", "int", base=36)
        0    13368
        dtype: int64

    This argument also controls the reverse conversion, from integer to string.

    .. doctest::

        >>> cast(5, "string", base=2)
        0    101
        dtype: string
        >>> cast(63, "string", base=8)
        0    77
        dtype: string
        >>> cast(1023, "string", base=16)
        0    3FF
        dtype: string
        >>> cast(13368, "string", base=36)
        0    ABC
        dtype: string
    """
    if not isinstance(val, int):
        raise TypeError(f"`base` must be an integer, not {repr(val)}")
    if val != 0 and not 2 <= val <= 36:
        raise ValueError(
            f"`base` must be 0 or >= 2 and <= 36, not {repr(val)}"
        )
    return val


@cast.argument(default=defaults["call"])
def call(val: Callable | None, context: dict) -> Callable:
    """Apply a callable over the input data, producing the desired output.

    This is only used for conversions from
    :class:`ObjectType <pdcast.ObjectType>`.  It allows users to specify a
    custom endpoint to perform this conversion, rather than relying exclusively
    on special methods (which is the default).
    
    TODO
    """
    if val is not None and not callable(val):
        raise TypeError(f"`call` must be callable, not {val}")
    return val


@cast.argument(default=defaults["downcast"])
def downcast(
    val: bool | type_specifier,
    context: dict
) -> types.CompositeType:
    """Losslessly reduce the precision of numeric data after converting.
    
    TODO
    """
    if val is None:
        return val
    if isinstance(val, bool):  # empty set is truthy, `None` is falsy
        return types.CompositeType() if val else None
    return resolve_type([val])


@cast.argument(default=defaults["errors"])
def errors(val: str, context: dict) -> str:
    """The rule to apply if/when errors are encountered during conversion.
    
    TODO
    """
    if val not in valid_errors:
        raise ValueError(
            f"`errors` must be one of {valid_errors}, not {repr(val)}"
        )
    return val



from cpython cimport datetime
from .epoch cimport Epoch


# constants
cdef dict timedelta_regex


# functions
cpdef object pandas_timedelta_to_ns(object delta)
cpdef object pytimedelta_to_ns(datetime.timedelta delta)
cpdef object numpy_timedelta64_to_ns(
    object delta,
    Epoch since,
    str unit = *,
    long int step_size = *
)



"""This module contains dispatched cast() implementations for boolean data."""
# pylint: disable=unused-argument
import pandas as pd

from pdcast import types

from .base import cast


@cast.overload("bool", "decimal")
def boolean_to_decimal(
    series: pd.Series,
    dtype: types.VectorType,
    **unused
) -> pd.Series:
    """Convert boolean data to a decimal data type."""
    target = dtype.dtype
    if isinstance(target, types.ObjectDtype):
        series = series + dtype.type_def(0)  # ~2x faster than loop
    return series.astype(target)


@cast.overload("bool", "datetime")
def boolean_to_datetime(
    series: pd.Series,
    dtype: types.VectorType,
    **unused
) -> pd.Series:
    """Convert boolean data to a datetime format."""
    # 2-step conversion: bool -> int, int -> datetime
    series = cast(series, "int", downcast=False, errors="raise")
    return cast(series, dtype, **unused)


@cast.overload("bool", "timedelta")
def boolean_to_timedelta(
    series: pd.Series,
    dtype: types.VectorType,
    **unused
) -> pd.Series:
    """Convert boolean data to a timedelta format."""
    # 2-step conversion: bool -> int, int -> timedelta
    series = cast(series, "int", downcast=False, errors="raise")
    return cast(series, dtype, **unused)



"""This module contains dispatched cast() implementations for data stored as
raw Python objects.
"""
# pylint: disable=unused-argument
from typing import Callable

import pandas as pd

from pdcast import types
from pdcast.resolve import resolve_type
from pdcast.detect import detect_type
from pdcast.util.vector import apply_with_errors

from .base import cast, generic_to_object, safe_apply


@cast.overload("object", "bool")
def object_to_boolean(
    series: pd.Series,
    dtype: types.ScalarType,
    call: Callable,
    errors: str,
    **unused
) -> pd.Series:
    """Convert unstructured objects to a boolean data type."""
    # if no callable is given, hook into object's __bool__ dunder
    use_dunder = call is None or call is bool
    if use_dunder:
        call = bool

    # apply callable
    series = apply_with_errors(series, call, errors=errors)
    if use_dunder:
        series = series.astype(resolve_type(bool).dtype)
    else:
        output_type = detect_type(series)
        if not types.BooleanType.contains(output_type):
            raise TypeError(
                f"`call` must produce boolean output, not {str(output_type)}"
            )
        series = series.astype(output_type.dtype)

    # convert bool[python] -> final repr
    return cast(series, dtype, call=call, errors=errors, **unused)


@cast.overload("object", "int")
def object_to_integer(
    series: pd.Series,
    dtype: types.ScalarType,
    call: Callable,
    errors: str,
    **unused
) -> pd.Series:
    """Convert unstructured objects to an integer data type."""
    # if no callable is given, hook into object's __int__ dunder
    use_dunder = call is None or call is int
    if use_dunder:
        call = int

    # apply callable
    series = apply_with_errors(series, call, errors=errors)
    if use_dunder:
        series = series.astype(resolve_type(int).dtype)
    else:
        output_type = detect_type(series)
        if not types.IntegerType.contains(output_type):
            raise TypeError(
                f"`call` must produce integer output, not {str(output_type)}"
            )
        series = series.astype(output_type.dtype)

    # convert int[python] -> final repr
    return cast(series, dtype, call=call, errors=errors, **unused)


@cast.overload("object", "float")
def object_to_float(
    series: pd.Series,
    dtype: types.ScalarType,
    call: Callable,
    errors: str,
    **unused
) -> pd.Series:
    """Convert unstructured objects to a float data type."""
    # if no callable is given, hook into object's __float__ dunder
    use_dunder = call is None or call is float
    if use_dunder:
        call = float

    # apply callable
    series = apply_with_errors(series, call, errors=errors)
    if use_dunder:
        series = series.astype(resolve_type(float).dtype)
    else:
        output_type = detect_type(series)
        if not types.FloatType.contains(output_type):
            raise TypeError(
                f"`call` must produce float output, not {str(output_type)}"
            )
        series = series.astype(output_type.dtype)

    # convert float[python] -> final repr
    return cast(series, dtype, call=call, errors=errors, **unused)


@cast.overload("object", "complex")
def object_to_complex(
    series: pd.Series,
    dtype: types.ScalarType,
    call: Callable,
    errors: str,
    **unused
) -> pd.Series:
    """Convert unstructured objects to a complex data type."""
    # if no callable is given, hook into object's __complex__ dunder
    use_dunder = call is None or call is complex
    if use_dunder:
        call = complex

    # apply callable
    series = apply_with_errors(series, call, errors=errors)
    if use_dunder:
        series = series.astype(resolve_type(complex).dtype)
    else:
        output_type = detect_type(series)
        if not types.ComplexType.contains(output_type):
            raise TypeError(
                f"`call` must produce complex output, not {str(output_type)}"
            )
        series = series.astype(output_type.dtype)

    # convert complex[python] -> final repr
    return cast(series, dtype, call=call, errors=errors, **unused)


@cast.overload("object", "decimal")
def object_to_decimal(
    series: pd.Series,
    dtype: types.ScalarType,
    call: Callable,
    errors: str,
    **unused
) -> pd.Series:
    """Convert unstructured objects to a decimal data type."""
    # if no callable is given, hook into object's __float__ dunder
    if call is None:
        call = lambda x: dtype.type_def(float(x))

    # check callable output matches dtype at every index
    return safe_apply(series=series, dtype=dtype, call=call, errors=errors)


@cast.overload("object", "datetime")
def object_to_datetime(
    series: pd.Series,
    dtype: types.ScalarType,
    call: Callable,
    errors: str,
    **unused
) -> pd.Series:
    """Convert unstructured objects to a datetime data type."""
    # if callable is given, convert directly
    if call:
        return safe_apply(series=series, dtype=dtype, call=call, errors=errors)

    # 2-step conversion: object -> float, float -> datetime
    series = cast(series, float, call=float, errors=errors)
    return cast(series, dtype, call=call, errors=errors, **unused)


@cast.overload("object", "timedelta")
def object_to_timedelta(
    series: pd.Series,
    dtype: types.ScalarType,
    call: Callable,
    errors: str,
    **unused
) -> pd.Series:
    """Convert unstructured objects to a timedelta data type."""
    # if callable is given, convert directly
    if call:
        return safe_apply(
            series=series,
            dtype=dtype,
            call=call,
            errors=errors
        )

    # 2-step conversion: object -> float, float -> timedelta
    series = cast(series, float, call=float, errors=errors)
    return cast(series, dtype, call=call, errors=errors, **unused)


@cast.overload("object", "string")
def object_to_string(
    series: pd.Series,
    dtype: types.ScalarType,
    call: Callable,
    errors: str,
    **unused
) -> pd.Series:
    """Convert unstructured objects to a string data type."""
    # if no callable is given, hook into object's __str__ dunder
    if call is None:
        if dtype.type_def is str:
            call = str
        else:
            call = lambda x: dtype.type_def(str(x))

    # check callable output matches dtype at every index
    return safe_apply(series=series, dtype=dtype, call=call, errors=errors)



cimport numpy as np

# constants
cdef unsigned int days_per_400_years
cdef unsigned short days_per_100_years
cdef unsigned short days_per_4_years
cdef unsigned short days_per_year
cdef np.ndarray days_per_month


# functions
cpdef object date_to_days(object year, object month, object day)
cpdef object days_in_month(object month, object year = *)
cpdef dict days_to_date(object days)
cpdef object is_leap_year(object year)
cpdef object leaps_between(object lower, object upper)



"""This module contains dispatched cast() implementations for decimal data."""
# pylint: disable=unused-argument
import datetime

import numpy as np
import pandas as pd

from pdcast import types
from pdcast.detect import detect_type
from pdcast.resolve import resolve_type
from pdcast.util import time
from pdcast.util.error import shorten_list
from pdcast.util.numeric import boundscheck, downcast_float, isinf, within_tol
from pdcast.util.round import Tolerance
from pdcast.util.vector import apply_with_errors

from .base import (
    cast, generic_to_boolean, generic_to_integer, snap_round
)


# TODO: pdcast.cast(pdcast.cast([1., 2., np.inf], "decimal"), "float")
# OverflowError: cannot convert Infinity to integer

# -> insert series.min == inf checks


@cast.overload("decimal", "bool")
def decimal_to_boolean(
    series: pd.Series,
    dtype: types.ScalarType,
    rounding: str,
    tol: Tolerance,
    errors: str,
    **unused
) -> pd.Series:
    """Convert decimal data to a boolean data type."""
    series = snap_round(
        series,
        tol=tol.real,
        rule=rounding,
        errors=errors
    )
    series = boundscheck(series, dtype=dtype, tol=tol.real, errors=errors)
    return generic_to_boolean(series, dtype, errors=errors)


@cast.overload("decimal", "integer")
def decimal_to_integer(
    series: pd.Series,
    dtype: types.ScalarType,
    rounding: str,
    tol: Tolerance,
    downcast: types.CompositeType,
    errors: str,
    **unused
) -> pd.Series:
    """Convert decimal data to an integer data type."""
    series = snap_round(
        series,
        tol=tol.real,
        rule=rounding,
        errors=errors
    )
    series = boundscheck(series, dtype=dtype, tol=tol.real, errors=errors)
    return generic_to_integer(
        series,
        dtype,
        tol=tol,
        downcast=downcast,
        errors=errors
    )


@cast.overload("decimal", "float")
def decimal_to_float(
    series: pd.Series,
    dtype: types.ScalarType,
    tol: Tolerance,
    downcast: types.CompositeType,
    errors: str,
    **unused
) -> pd.Series:
    """Convert decimal data to a floating point data type."""
    # do naive conversion
    if dtype.itemsize > 8:
        # NOTE: series.astype() implicitly calls Decimal.__float__(), which
        # is limited to 64-bits.  Converting to an intermediate string
        # representation avoids this and maintains full precision
        result = series.astype(str).astype(dtype)
    else:
        result = series.astype(dtype)

    # check for overflow
    if int(series.min()) < dtype.min or int(series.max()) > dtype.max:
        infs = isinf(result) ^ isinf(series)
        if infs.any():
            if errors == "coerce":
                result = result[~infs]
                series = series[~infs]  # mirror on original
            else:
                raise OverflowError(
                    f"values exceed {dtype} range at index "
                    f"{shorten_list(series[infs].index.values)}"
                )

    # backtrack to check for precision loss
    if errors != "coerce":  # coercion ignores precision loss
        series_type = detect_type(series)
        bad = ~within_tol(
            series,
            cast(result, dtype=series_type, errors="raise"),
            tol=tol.real
        )
        if bad.any():
            raise ValueError(
                f"precision loss exceeds tolerance {float(tol.real):g} at "
                f"index {shorten_list(series[bad].index.values)}"
            )

    if downcast is not None:
        return downcast_float(result, tol=tol, smallest=downcast)
    return result


@cast.overload("decimal", "complex")
def decimal_to_complex(
    series: pd.Series,
    dtype: types.ScalarType,
    tol: Tolerance,
    downcast: types.CompositeType,
    errors: str,
    **unused
) -> pd.Series:
    """Convert decimal data to a complex data type."""
    # 2-step conversion: decimal -> float, float -> complex
    series = cast(
        series,
        "float",
        tol=tol,
        downcast=None,
        errors=errors
    )
    return cast(
        series,
        dtype,
        tol=tol,
        downcast=downcast,
        errors=errors,
        **unused
    )


@cast.overload("decimal", "decimal")
def decimal_to_decimal(
    series: pd.Series,
    dtype: types.ScalarType,
    errors: str,
    **unused
) -> pd.Series:
    """Convert boolean data to a decimal data type."""
    # trivial case
    if detect_type(series) == dtype:
        return series

    target = dtype.dtype
    if isinstance(dtype.dtype, types.ObjectDtype):
        series = apply_with_errors(series, dtype.type_def, errors=errors)
    return series.astype(target)


@cast.overload("decimal", "datetime")
def decimal_to_datetime(
    series: pd.Series,
    dtype: types.ScalarType,
    unit: str,
    step_size: int,
    since: time.Epoch,
    tz: datetime.tzinfo,
    tol: Tolerance,
    errors: str,
    **unused
) -> pd.Series:
    """Convert integer data to a datetime data type."""
    # 2-step conversion: decimal -> ns, ns -> datetime
    series = to_ns(series, unit=unit, step_size=step_size, since=since)

    # account for non-utc epoch
    if since:
        series += since.offset

    # check for overflow and upcast if applicable
    series = boundscheck(series, dtype=dtype, tol=tol.real, errors=errors)

    # convert to final representation
    return cast(
        series,
        dtype,
        unit="ns",
        step_size=1,
        since=time.Epoch("utc"),
        tz=tz,
        errors=errors,
        **unused
    )


@cast.overload("decimal", "timedelta")
def decimal_to_timedelta(
    series: pd.Series,
    dtype: types.ScalarType,
    unit: str,
    step_size: int,
    since: time.Epoch,
    tol: Tolerance,
    errors: str,
    **unused
) -> pd.Series:
    """Convert integer data to a timedelta data type."""
    # 2-step conversion: decimal -> ns, ns -> timedelta
    series = to_ns(series, unit=unit, step_size=step_size, since=since)

    # check for overflow and upcast if necessary
    series = boundscheck(series, dtype=dtype, tol=tol.real, errors=errors)

    # convert to final representation
    return cast(
        series,
        dtype,
        unit="ns",
        step_size=1,
        since=since,
        errors=errors,
        **unused,
    )


#######################
####    PRIVATE    ####
#######################


def to_ns(
    series: pd.Series,
    unit: str,
    step_size: int,
    since: time.Epoch
) -> pd.Series:
    """Convert an integer number of time units into nanoseconds from a given
    epoch.
    """
    # round fractional inputs to the nearest nanosecond
    if unit == "Y":
        result = time.round_years_to_ns(series * step_size, since=since)
    elif unit == "M":
        result = time.round_months_to_ns(series * step_size, since=since)
    else:
        as_pyint = np.frompyfunc(int, 1, 1)
        result = series
        if step_size != 1:
            result *= step_size
        if unit != "ns":
            result *= time.as_ns[unit]
        result = as_pyint(result)

    return result.astype(resolve_type(int).dtype)



"""This module describes a single function (``round_float()``), which performs
customizable, vectorized rounding on floating point numbers.
"""
cimport numpy as np
import numpy as np
import pandas as pd


######################
####    PUBLIC    ####
######################


cpdef object round_float(
    object val,
    short decimals,
    str rule
):
    """Round a float or array of floats according to the specified `rule`.

    Parameters
    ----------
    val : float | array_like
        The value to be rounded.  Can be vectorized.
    rule : str, default 'half_even'
        A string specifying the rounding strategy to use.  Must be one of
        ('floor', 'ceiling', 'down', 'up', 'half_floor', 'half_ceiling',
        'half_down', 'half_up', 'half_even'), where `up/down` round
        away/toward zero, and `ceiling/floor` round toward +/- infinity,
        respectively.
    decimals : int, default 0
        The number of decimals to round to.  Positive numbers count to the
        right of the decimal point, and negative values count to the left.
        0 represents rounding in the ones place of `val`.  This follows the
        convention set out in `numpy.around()`.

    Returns
    -------
    float | array_like
        The result of rounding `val` according to the given rule.

    Raises
    ------
    ValueError
        If `rule` is not one of the accepted rounding rules ('floor',
        'ceiling', 'down', 'up', 'half_floor', 'half_ceiling', 'half_down',
        'half_up', 'half_even').
    """
    cdef object round_func
    cdef str err_msg
    cdef object scale_factor

    # select rounding strategy
    try:
        # use numpy-accelerated rounding if available, else default to generic
        if getattr(val, "dtype", np.dtype("O")).kind == "O":
            round_func = generic_rounding_rules[rule]
        else:
            round_func = specialized_rounding_rules[rule]
    except KeyError as err:
        err_msg = (
            f"`rule` must be one of {tuple(generic_rounding_rules)}, not "
            f"{repr(rule)}"
        )
        raise ValueError(err_msg) from err

    if decimals:
        scale_factor = 10 ** decimals
        return round_func(val * scale_factor) / scale_factor

    return round_func(val)


#######################
####    PRIVATE    ####
#######################


cdef object _generic_round_half_even(object v):
    cdef object floor_even = (v // 1 % 2) * 2 - 1
    return floor_even * ((floor_even * v + 0.5) // 1)


cdef dict generic_rounding_rules = {
    "floor": lambda v: v // 1,
    "ceiling": lambda v: -(-v // 1),
    "down": lambda v: abs(v) // 1 * np.sign(v),
    "up": lambda v: -(-abs(v) // 1) * np.sign(v),
    "half_floor": lambda v: -((-v + 0.5) // 1),
    "half_ceiling": lambda v: (v + 0.5) // 1,
    "half_down": lambda v: -((-abs(v) + 0.5) // 1) * np.sign(v),
    "half_up": lambda v: (abs(v) + 0.5) // 1 * np.sign(v),
    "half_even": _generic_round_half_even
}


cdef dict specialized_rounding_rules = {
    "floor": np.floor,
    "ceiling": np.ceil,
    "down": np.trunc,
    "up": lambda x: np.sign(x) * np.ceil(np.abs(x)),
    "half_floor": lambda x: np.ceil(x - 0.5),
    "half_ceiling": lambda x: np.floor(x + 0.5),
    "half_down": lambda x: np.sign(x) * np.ceil(np.abs(x) - 0.5),
    "half_up": lambda x: np.sign(x) * np.floor(np.abs(x) + 0.5),
    "half_even": np.round
}



"""This package contains various utilities related to ``pdcast`` functionality,
including rounding, datetime manipulations, type hints, and basic data
structures.

Subpackages
-----------
array
    Automatic generation of ``ExtensionDtypes`` for arbitrary data, as well as
    specialized extensions for specific data types.

round
    Customizable rounding for numeric data, with a wider range of options than
    numpy or pandas.

structs
    Basic data structures that are generally applicable and can be used as
    recipes for other projects.

time
    Calendar-accurate datetime manipulations and string conversions with
    infinite range.

Modules
-------
error
    utilities for handling errors raised by ``pdcast`` functions and methods.

type_hints
    type hints for mypy and other static type checkers.
"""


"""This module provides calendar-accurate unit conversions for datetime and
timedelta data.

The units covered in this module are the same as those implemented in the
``numpy.datetime64``/``numpy.timedelta64`` interface.

Functions
---------
convert_unit
    Arbitrary unit conversions.

round_years_to_ns
    Round a vector of (possibly fractional) years to the nearest nanosecond.

round_months_to_ns
    Round a vector of (possibly fractional) months to the nearest nanosecond.

Constants
---------
as_ns
    Conversion factors from each of the regular units (i.e. not "M" or "Y") to
    nanoseconds.

valid_units
    A list of units that are recognized by this package.
"""
import datetime
import decimal

import numpy as np
cimport numpy as np
import pandas as pd

from pdcast.util.round import round_div
cimport pdcast.util.time.epoch as epoch
import pdcast.util.time.epoch as epoch
from pdcast.util.type_hints import array_like, datetime_like, numeric

from .calendar import (
    date_to_days, days_in_month, days_to_date, is_leap_year
)


#########################
####    CONSTANTS    ####
#########################


cdef dict as_ns = {
    # "as": 1e-9,
    # "fs": 1e-6,
    # "ps": 1e-3,
    "ns": 1,
    "us": 10**3,
    "ms": 10**6,
    "s": 10**9,
    "m": 60 * 10**9,
    "h": 60 * 60 * 10**9,
    "D": 24 * 60 * 60 * 10**9,
    "W": 7 * 24 * 60 * 60 * 10**9,
}
as_ns_public = as_ns  # python-facing alias for as_ns


cdef tuple valid_units = tuple(as_ns) + ("M", "Y")
valid_units_public = valid_units  # python-facing alias for valid_units


cdef object cast_to_int = np.frompyfunc(int, 1, 1)


######################
####    PUBLIC    ####
######################


cpdef object convert_unit(
    object val,
    str from_unit,
    str to_unit,
    str rounding = "down",
    epoch.Epoch since = None
):
    """Convert integer quantities of a given time unit into a different unit.

    This function performs the same operation as :func:`convert_unit_float()`,
    except that it is designed for integer unit representations and avoids
    conversion to imprecise floating-point formats.

    Parameters
    ----------
    arg : int | array-like
        The quantity to convert.  Can be vectorized.
    from_unit : {'ns', 'us', 'ms', 's', 'm', 'h', 'D', 'W', 'M', 'Y'}
        The unit to convert from.
    to_unit : {'ns', 'us', 'ms', 's', 'm', 'h', 'D', 'W', 'M', 'Y'}
        The unit to convert to.
    rounding : {'floor', 'ceiling', 'down', 'up', 'half_floor', 'half_ceiling',
    'half_down', 'half_up', 'half_even', None}
        The rounding strategy to use in the case of residual units.  If this
        is set to `None`, then the result will be coerced to float and the
        residual appended as a decimal component.  Otherwise, it will stay in
        a pure-integer representation, and the specified rounding rule will be
        applied on that basis.
    since : datetime_like, default pd.Timestamp("2001-01-01 00:00:00+0000")
        The date from which to begin counting.  This is only used when
        converting to or from units 'M' and 'Y', in order to accurately account
        for leap days and unequal month lengths.  Only the `year`, `month`, and
        `day` components are used.  Defaults to '2001-01-01 00:00:00+0000',
        which represents the start of a 400-year Gregorian calendar cycle.

    Returns
    -------
    int | float | array-like
        The result of the unit conversion.  If `rounding=None`, this will
        be coerced to float, with residuals as decimal components.  Otherwise,
        it will stay in integer format and apply the given rounding rule.

    Raises
    ------
    ValueError
        If either `from_unit` or `to_unit` is not one of the recognized
        units ('ns', 'us', 'ms', 's', 'm', 'h', 'D', 'W', 'M', 'Y').

    See Also
    --------
    convert_unit_float : convert fractional numbers of units.
    ns_to_datetime : convert nanosecond UTC offsets into datetimes.
    ns_to_timedelta : convert nanosecond UTC offsets into timedeltas.

    Examples
    --------
    Units can be regular ('ns', 'us', 'ms', 's', 'm', 'h', 'D', 'W'):

    >>> convert_unit_integer(1, "s", "ns")
    1000000000
    >>> convert_unit_integer(123, "s", "ns")
    123000000000
    >>> convert_unit_integer(10, "D", "s")
    864000

    Or irregular ('M', 'Y'):

    >>> convert_unit_integer(1, "Y", "M")
    12
    >>> convert_unit_integer(42, "M", "Y")
    3

    With conversion between the two:

    >>> convert_unit_integer(1, "Y", "D")
    365
    >>> convert_unit_integer(1, "Y", "D", since=pd.Timestamp("2000-01-01"))
    366
    >>> convert_unit_integer(32, "M", "s")
    84067200
    >>> convert_unit_integer(365, "D", "Y")
    1
    >>> convert_unit_integer(365, "D", "Y", since=pd.Timestamp("2000-01-01"))
    0

    Optionally, results can be returned as floats, exposing the residuals:

    >>> convert_unit_integer(1, "s", "m", rounding=None)
    0.016666666666666666
    >>> convert_unit_integer(365, "D", "Y", rounding=None, since=pd.Timestamp("2000-01-01"))
    0.9972677595628415
    """
    # ensure units are valid
    if from_unit not in valid_units:
        raise ValueError(
            f"`from_unit` must be one of {valid_units}, not {repr(from_unit)}"
        )
    if to_unit not in valid_units:
        raise ValueError(
            f"`to_unit` must be one of {valid_units}, not {repr(to_unit)}"
        )
    if since is None:
        since = epoch.Epoch("utc")

    # trivial case - no conversion necessary
    if from_unit == to_unit:
        return val

    # regular -> regular ('ns', 'us', 'ms', 's', 'm', 'h', 'D', 'W')
    if from_unit in as_ns and to_unit in as_ns:
        if rounding is None:
            return val * as_ns[from_unit] / as_ns[to_unit]
        return round_div(
            val * as_ns[from_unit],
            as_ns[to_unit],
            rule=rounding
        )

    # irregular -> irregular ('M', 'Y')
    if from_unit not in as_ns and to_unit not in as_ns:
        # month -> year
        if from_unit == "M" and to_unit == "Y":
            if rounding is None:
                return val / 12
            return round_div(val, 12, rule=rounding)

        # year -> month
        return 12 * val

    # regular -> irregular
    if from_unit in as_ns:
        return _convert_unit_regular_to_irregular(
            val,
            from_unit=from_unit,
            to_unit=to_unit,
            rounding=rounding,
            since=since
        )

    # irregular -> regular
    return _convert_unit_irregular_to_regular(
        val,
        from_unit=from_unit,
        to_unit=to_unit,
        rounding=rounding,
        since=since
    )


#######################
####    PRIVATE    ####
#######################


cpdef object round_years_to_ns(object years, epoch.Epoch since):
    """Convert a fractional number of years from a given epoch into an integer
    number of nanoseconds.
    """
    cdef object diff
    cdef object residual
    cdef object end
    cdef object unit_length

    # handle integer and fractional components separately
    diff = cast_to_int(years)
    residual = years - diff

    # convert integer component to utc day offset and establish ending date
    diff = date_to_days(since.year + diff, since.month, since.day)
    end = days_to_date(diff)

    # establish length of final unit in days.  NOTE: if start date occurs on or
    # after a potential leap day, use next year's calendar.  If residual years
    # are negative, use last year's instead.
    unit_length = 365 + is_leap_year(
        end["year"] + (since.ordinal > 58) - (residual < 0)
    )

    # subtract off epoch to get total elapsed days
    if since:
        diff -= since.offset // as_ns["D"]

    # scale fractional component by unit length and reintroduce integer
    return diff * as_ns["D"] + cast_to_int(residual * unit_length * as_ns["D"])


cpdef object round_months_to_ns(object months, epoch.Epoch since):
    """Convert a fractional number of months from a given epoch into an integer
    number of nanoseconds.
    """
    cdef object diff
    cdef object residual
    cdef object end
    cdef object unit_length

    # handle integer and fractional components separately
    diff = cast_to_int(months)
    residual = months - diff

    # convert integer component to utc day offset and establish ending date
    diff = date_to_days(since.year, since.month + diff, since.day)
    end = days_to_date(diff)

    # establish length of final unit in days.  NOTE: if residual months are
    # negative, use the previous month.
    unit_length = days_in_month(end["month"] - (residual < 0), end["year"])

    # subtract off start date to get total elapsed days
    if since:
        diff -= since.offset // as_ns["D"]

    # scale fractional component by unit length and reintroduce integer
    return diff * as_ns["D"] + cast_to_int(residual * unit_length * as_ns["D"])


cdef object _convert_unit_irregular_to_regular(
    object val,
    str from_unit,
    str to_unit,
    str rounding,
    epoch.Epoch since
):
    """Helper to convert integer numbers of irregular units ('M', 'Y') to
    regular units ('ns', 'us', 'ms', 's', 'm', 'h', 'D', 'W').
    """
    # month -> day
    if from_unit == "M":
        val = date_to_days(
            since.year,
            since.month + val,
            since.day
        )

    # year -> day
    else:
        val = date_to_days(
            since.year + val,
            since.month,
            since.day
        )

    # move origin to epoch
    if since:
        val -= since.offset // as_ns["D"]

    # day -> day
    if to_unit == "D":
        return val

    # day -> week
    if to_unit == "W":
        if rounding is None:
            return val / 7
        return round_div(val, 7, rule=rounding)

    # day -> ns, us, ms, s, m, h
    val *= as_ns["D"]
    if rounding is None:
        return val / as_ns[to_unit]
    return round_div(val, as_ns[to_unit], rule=rounding)


cdef object _convert_unit_regular_to_irregular(
    object val,
    str from_unit,
    str to_unit,
    str rounding,
    epoch.Epoch since
):
    """Helper to convert integer numbers of regular units ('ns', 'us', 'ms',
    's', 'm', 'h', 'D', 'W') to irregular units ('M', 'Y').
    """
    cdef object diff
    cdef object residual
    cdef object end
    cdef object unit_length

    # convert val to whole days + residual ns
    val = val * as_ns[from_unit]
    diff = round_div(val, as_ns["D"], rule="down")
    residual = val - diff * as_ns["D"]

    # convert whole days into final end date
    if since:
        diff += since.offset // as_ns["D"]  # move origin to epoch
    end = days_to_date(diff)

    # build result, starting with whole number of years
    result = end["year"] - since.year

    # day -> year
    if to_unit == "Y":
        # move origin to nearest multiple of `since`
        diff -= date_to_days(end["year"], since.month, since.day)

        # get length of final unit.  NOTE: if `since` occurs on or after a
        # leap day, use next year's calendar.
        unit_length = 365 + is_leap_year(end["year"] + (since.ordinal > 58))

    # day -> month
    else:
        # get whole number of months
        result *= 12
        result += end["month"] - since.month

        # move origin to nearest multiple of `since`
        diff -= date_to_days(end["year"], end["month"], since.day)

        # establish length of final month
        unit_length = days_in_month(end["month"], end["year"])

    # convert back to nanoseconds and reintroduce residuals
    unit_length *= as_ns["D"]
    diff *= as_ns["D"]
    diff += residual

    # `diff` now holds the time separation (in ns) of `result` from the
    # nearest occurence of `since`, bounded to within 1 unit.  Negative
    # values indicate that `end` occurs before `since`, meaning `result`
    # is an overestimate.  Positive values indicate the opposite.  This
    # allows us to apply customizable rounding while retaining full integer
    # accuracy.

    # return exact (do not round)
    if rounding is None:
        return result + diff / unit_length  # float division

    # round toward/away from zero
    if rounding == "floor":
        return result - (diff < 0)
    if rounding == "ceiling":
        return result + (diff > 0)
    if rounding == "down":
        result -= ((val > 0) & (diff < 0))  # floor where val > 0
        result += ((val < 0) & (diff > 0))  # ceiling where val < 0
        return result
    if rounding == "up":
        result -= ((val < 0) & (diff < 0))  # floor where val < 0
        result += ((val > 0) & (diff > 0))  # ceiling where val > 0
        return result

    # corrections for round to nearest
    if to_unit == "M":
        if rounding == "half_down":
            diff += (val < 0)
        elif rounding == "half_up":
            diff -= (val < 0)

    # round half even has to be handled separately
    if rounding == "half_even":
        result += np.where(
            result % 2,
            round_div(diff, unit_length, rule="half_ceiling"),
            round_div(diff, unit_length, rule="half_floor")
        )
        return result[()] if not result.shape else result

    # round to nearest
    return result + round_div(diff, unit_length, rule=rounding)


"""This module describes a ``NullType`` object, which can be used to represent
missing values.
"""
import pandas as pd

from .base import Type


class Missing(Type, backend="missing"):
    """Missing value type."""

    aliases = {
        "null", "none", "None", "NaN", "nan", "NA", "na", "missing",
        type(None), type(pd.NA)
    }
    scalar = type(pd.NA)
    missing = pd.NA



# pylint: disable=redefined-outer-name, unused-argument
import pandas as pd

from pdcast.decorators.attachable import attachable, VirtualAttribute
from pdcast.decorators.dispatch import dispatch
from pdcast.decorators.extension import extension_func
from pdcast.util.numeric import real, imag
from pdcast.util.round import round_decimal, round_div, round_float


######################
####    PUBLIC    ####
######################


@attachable
@extension_func
@dispatch("series")
def round(
    series: pd.Series,
    decimals: int = 0,
    rule: str = "half_even"
) -> pd.Series:
    """TODO: copy from abstract docs"""
    if rule != "half_even":
        raise ValueError(
            f"original pandas `Series.round()` implementation accepts only "
            f"rule='half_even', not {repr(rule)}"
        )

    original = getattr(series.round, "original", series.round)
    return original(decimals=decimals)


#########################
####    ARGUMENTS    ####
#########################


valid_rules = (
    "floor", "ceiling", "down", "up", "half_floor", "half_ceiling",
    "half_down", "half_up", "half_even"
)


@round.argument
def decimals(val: int, state: dict) -> int:
    """Ensure that `decimals` are integer-like."""
    return int(val)


@round.argument
def rule(val: str | None, state: dict) -> str:
    """The rounding rule to use for numeric conversions.

    Parameters
    ----------
    val : str | None
        An optional string specifying the rounding rule to use, or :data:`None`
        to indicate that no rounding will be applied.  Defaults to :data:`None`.

    Returns
    -------
    str | None
        A validated version of the string passed to ``val`` or :data:`None`.

    Raises
    ------
    TypeError
        If ``val`` is not a string or :data:`None <python:None>`.
    ValueError
        If ``val`` does not correspond to one of the recognized rounding rules.

    Notes
    -----
    The available options for this argument are as follows:

        *   ``None`` - do not round.
        *   ``"floor"`` - round toward negative infinity.
        *   ``"ceiling"`` - round toward positive infinity.
        *   ``"down"`` - round toward zero.
        *   ``"up"`` - round away from zero.
        *   ``"half_floor"`` - round to nearest with ties toward positive infinity.
        *   ``"half_ceiling"`` - round to nearest with ties toward negative
            infinity.
        *   ``"half_down"`` - round to nearest with ties toward zero.
        *   ``"half_up"`` - round to nearest with ties away from zero.
        *   ``"half_even"`` - round to nearest with ties toward the `nearest even
            value <https://en.wikipedia.org/wiki/Rounding#Rounding_half_to_even>`_.
            Also known as *convergent rounding*, *statistician's rounding*, or
            *banker's rounding*.

    This argument is applied **after**
    :func:`tol <pdcast.convert.arguments.tol>`.

    Examples
    --------
    .. doctest::

        >>> cast([-1.5, -0.5, 0.2, 1.7], "int", rounding="floor")
        0   -2
        1   -1
        2    0
        3    1
        dtype: int64
        >>> cast([-1.5, -0.5, 0.2, 1.7], "int", rounding="ceiling")
        0   -1
        1    0
        2    1
        3    2
        dtype: int64
        >>> cast([-1.5, -0.5, 0.2, 1.7], "int", rounding="down")
        0   -1
        1    0
        2    0
        3    1
        dtype: int64
        >>> cast([-1.5, -0.5, 0.2, 1.7], "int", rounding="up")
        0   -2
        1   -1
        2    1
        3    2
        dtype: int64
        >>> cast([-1.5, -0.5, 0.2, 1.7], "int", rounding="half_floor")
        0   -2
        1   -1
        2    0
        3    2
        dtype: int64
        >>> cast([-1.5, -0.5, 0.2, 1.7], "int", rounding="half_ceiling")
        0   -1
        1    0
        2    0
        3    2
        dtype: int64
        >>> cast([-1.5, -0.5, 0.2, 1.7], "int", rounding="half_down")
        0   -1
        1    0
        2    0
        3    2
        dtype: int64
        >>> cast([-1.5, -0.5, 0.2, 1.7], "int", rounding="half_up")
        0   -2
        1   -1
        2    0
        3    2
        dtype: int64
        >>> cast([-1.5, -0.5, 0.2, 1.7], "int", rounding="half_even")
        0   -2
        1    0
        2    0
        3    2
        dtype: int64
    """
    if not (val is None or isinstance(val, str)):
        raise TypeError(
            f"`rounding` must be a string or None, not {repr(val)}"
        )
    if val is not None and val not in valid_rules:
        raise ValueError(
            f"`rounding` must be one of {valid_rules}, not {repr(val)}"
        )
    return val


#######################
####    PRIVATE    ####
#######################


@round.overload("int")
def _round_integer(
    series: pd.Series,
    decimals: int,
    rule: str
) -> pd.Series:
    """Round an integer series to the given number of decimals.

    NOTE: this implementation does nothing unless the input to `decimals` is
    negative.
    """
    if decimals < 0:
        scale = 10**(-1 * decimals)
        return round_div(series, scale, rule=rule) * scale
    return series


@round.overload("decimal")
def _round_decimal(
    series: pd.Series,
    decimals: int,
    rule: str,
) -> pd.Series:
    """Overloaded round() implementation for decimal data."""
    return round_decimal(series, decimals=decimals, rule=rule)


@round.overload("float")
def _round_float(
    series: pd.Series,
    decimals: int,
    rule: str
) -> pd.Series:
    """Overloaded round() implementation for float data."""
    return round_float(series, decimals=decimals, rule=rule)


@round.overload("complex")
def _round_complex(
    series: pd.Series,
    decimals: int,
    rule: str
) -> pd.Series:
    """Round a complex series to the given number of decimal places using
    the specified rounding rule.
    """
    real_part = round_float(real(series), decimals=decimals, rule=rule)
    imag_part = round_float(imag(series), decimals=decimals, rule=rule)
    return real_part + imag_part * 1j



"""Mypy stubs for pdcast/util/round/integer.pyx"""

from pdcast.util.type_hints import array_like

def round_div(
    numerator: int | array_like, denominator: int | array_like, rule: str
) -> int | array_like:
    ...



"""Experimental support for numpy.datetime64-backed extension arrays.

It is not currently possible to support these due to limitations in the way
pandas formats repr() strings for M8 arrays.
"""
import numbers
from typing import Any, Iterator

import numpy as np
import pandas as pd
from pandas.api.extensions import register_extension_dtype
from pandas.core.arrays import ExtensionArray, ExtensionScalarOpsMixin
from pandas.core.dtypes.base import ExtensionDtype
from pandas.core.dtypes.generic import ABCDataFrame, ABCIndex, ABCSeries

import pdcast.convert as convert

import pdcast.util.time as time
from pdcast.util.type_hints import type_specifier


# NOTE: it is currently not possible to store np.datetime64 objects in a
# pandas series with units other than `1ns`.  Pandas internally formats these
# by checking the ExtensionDtype's `kind` and `type` attributes, which can
# be worked around.  What cannot be worked around is when pandas formats the
# underlying array itself, which we have no control over.  The only way to
# bypass this would be to change the behavior of np.asarray() on our
# ExtensionArrays or store them in a non-M8 container, which negates all the
# benefits of making an explicit ExtensionDtype in the first place.  As such,
# this module is currently dead in the water, though I haven't deleted it in
# case the internal pandas implementation changes in the future.


@register_extension_dtype
class M8Dtype(ExtensionDtype):
    """An ``ExtensionDtype`` that stores literal ``np.datetime64`` objects."""

    # NOTE: can't use kind="M" or type=np.datetime64 because it causes pandas
    # to auto-format results as pd.Timestamp objects during repr() calls, which
    # fails for any units other than 1ns.

    kind = "O"  # workaround for above
    type = type(None)  # workaround for above
    itemsize = 8
    na_value = np.datetime64("nat")
    _is_boolean = False
    _is_numeric = False
    _can_hold_na = True
    _metadata = ("unit", "step_size")

    def __init__(self, unit: str = "ns", step_size: int = 1):
        if unit not in time.valid_units:
            raise TypeError(
                f"unit must be one of {time.valid_units}, not {repr(unit)}"
            )
        if step_size < 1:
            raise TypeError(f"step_size must be >= 1, not {step_size}")
        self.unit = unit
        self.step_size = step_size

    @property
    def name(self) -> str:
        if self.step_size == 1:
            return f"M8[{self.unit}]"
        return f"M8[{self.step_size}{self.unit}]"

    @classmethod
    def construct_array_type(cls):
        return M8Array

    @classmethod
    def construct_from_string(cls: type, string: str) -> ExtensionDtype:
        """Construct this type from a string.

        This is useful mainly for data types that accept parameters.
        For example, a period dtype accepts a frequency parameter that
        can be set as ``period[H]`` (where H means hourly frequency).
        By default, in the abstract class, just the name of the type is
        expected. But subclasses can overwrite this method to accept
        parameters.

        Parameters
        ----------
        string : str
            The name of the type, for example ``category``.

        Returns
        -------
        ExtensionDtype
            Instance of the dtype.

        Raises
        ------
        TypeError
            If a class cannot be constructed from this 'string'.

        Examples
        --------
        For extension dtypes with arguments the following may be an
        adequate implementation.

        >>> import re
        >>> @classmethod
        ... def construct_from_string(cls, string):
        ...     pattern = re.compile(r"^my_type\[(?P<arg_name>.+)\]$")
        ...     match = pattern.match(string)
        ...     if match:
        ...         return cls(**match.groupdict())
        ...     else:
        ...         raise TypeError(
        ...             f"Cannot construct a '{cls.__name__}' from "
        ...             f"'{string}'"
        ...         )
        """
        if not isinstance(string, str):
            raise TypeError(
                f"'construct_from_string' expects a string, got {type(string)}"
            )

        # disable string construction - use ``pdcast.resolve_type()`` instead
        raise TypeError(f"Cannot construct a '{cls.__name__}' from '{string}'")

    def _get_common_dtype(self, dtypes: list):
        if len(set(dtypes)) == 1:  # only itself
            return self
        return np.dtype(f"M8[{self.step_size}{self.unit}]")

    def __repr__(self) -> str:
        return f"M8Dtype(unit={repr(self.unit)}, step_size={self.step_size})"


class M8Array(ExtensionArray, ExtensionScalarOpsMixin):
    """An ``ExtensionArray`` for storing literal ``np.datetime64`` objects."""

    __array_priority__ = 1000  # this is used in pandas test code

    def __init__(self, values, dtype=None, copy=False):
        self._data = np.asarray(values, dtype=dtype._get_common_dtype([]))

        # aliases for common attributes to ensure pandas support
        self._items = self.data = self._ndarray = self._data
        self._dtype = dtype

        self.tz = None

    @classmethod
    def _from_sequence(cls, scalars, dtype=None, copy=False):
        """Construct a new ExtensionArray from a sequence of scalars.

        Parameters
        ----------
        scalars : Sequence
            Each element will be an instance of the scalar type for this
            array, ``cls.dtype.type`` or be converted into this type in this
            method.
        dtype : dtype, optional
            Construct for this particular dtype.  This should be a Dtype
            compatible with the ExtensionArray.
        copy : bool, default False
            If True, copy the underlying data.

        Returns
        -------
        ExtensionArray
        """
        return cls(scalars, dtype=dtype, copy=copy)

    @classmethod
    def _from_factorized(cls, values, original):
        """Reconstruct an ExtensionArray after factorization.

        Parameters
        ----------
        values : ndarray
            An integer ndarray with the factorized values.
        original : ExtensionArray
            The original ExtensionArray that factorize was called on.

        See Also
        --------
        factorize : Top-level factorize method that dispatches here.
        ExtensionArray.factorize : Encode the array as an enumerated type.
        """
        return cls(values, dtype=original.dtype)

    def __getitem__(self, item):
        """Select a subset of self.

        Parameters
        ----------
        item : int, slice, or ndarray
            * int: The position in 'self' to get.
            * slice: A slice object, where 'start', 'stop', and 'step' are
              integers or None
            * ndarray: A 1-d boolean NumPy ndarray the same length as 'self'
            * list[int]:  A list of int

        Returns
        -------
        item : scalar or ExtensionArray

        Notes
        -----
        For scalar ``item``, return a scalar value suitable for the array's
        type. This should be an instance of ``self.dtype.type``.

        For slice ``key``, return an instance of ``ExtensionArray``, even
        if the slice is length 0 or 1.

        For a boolean mask, return an instance of ``ExtensionArray``, filtered
        to the values where ``item`` is True.
        """
        if isinstance(item, numbers.Integral):
            return self._data[item]
        return type(self)(self._data[item])

    def __setitem__(self, key, value):
        """Set one or more values inplace.

        This method is not required to satisfy the pandas extension array
        interface.

        Parameters
        ----------
        key : int, ndarray, or slice
            When called from, e.g. ``Series.__setitem__``, ``key`` will be
            one of
            * scalar int
            * ndarray of integers.
            * boolean ndarray
            * slice object
        value : ExtensionDtype.type, Sequence[ExtensionDtype.type], or object
            value or values to be set of ``key``.

        Returns
        -------
        None
        """
        if pd.api.types.is_list_like(value):
            if pd.api.types.is_scalar(key):
                raise ValueError(
                    "setting an array element with a sequence."
                )
            value = np.asarray(convert.cast(value, self._pdcast_type))
        # else:
        #     value = convert.cast(value, self._pdcast_type)[0]
        self._data[key] = value

    def __len__(self) -> int:
        """Length of this array

        Returns
        -------
        length : int
        """
        return len(self._data)

    def __iter__(self) -> Iterator[Any]:
        """Iterate over elements of the array."""
        return iter(self._data)

    def __eq__(self, other: type_specifier):
        """Return for `self == other` (element-wise equality)."""
        if isinstance(other, (pd.Series, pd.Index, pd.DataFrame)):
            return NotImplemented
        return self._data == other

    # def __repr__(self) -> str:
    #     if self.ndim > 1:
    #         return self._repr_2d()

    #     return "hello world"


    # def _formatter(self, boxed: bool = False) -> Callable[[Any], str | None]:
    #     """Formatting function for scalar values.

    #     This is used in the default '__repr__'. The returned formatting
    #     function receives instances of your scalar type.

    #     Parameters
    #     ----------
    #     boxed : bool, default False
    #         An indicated for whether or not your array is being printed
    #         within a Series, DataFrame, or Index (True), or just by
    #         itself (False). This may be useful if you want scalar values
    #         to appear differently within a Series versus on its own (e.g.
    #         quoted or not).

    #     Returns
    #     -------
    #     Callable[[Any], str]
    #         A callable that gets instances of the scalar type and
    #         returns a string. By default, :func:`repr` is used
    #         when ``boxed=False`` and :func:`str` is used when
    #         ``boxed=True``.
    #     """
    #     return lambda *args, **kwargs: "hello world"

    @property
    def dtype(self):
        """An instance of 'ExtensionDtype'."""
        return self._dtype

    @property
    def nbytes(self):
        """The number of bytes needed to store this object in memory."""
        return self.dtype.itemsize * len(self)

    def astype(self, dtype, copy: bool = True):
        """Cast to a NumPy array or ExtensionArray with 'dtype'.

        Parameters
        ----------
        dtype : str or dtype
            Typecode or data-type to which the array is cast.
        copy : bool, default True
            Whether to copy the data, even if not necessary. If False, a copy
            is made only if the old dtype does not match the new dtype.

        Returns
        -------
        array : np.ndarray or ExtensionArray
            An ExtensionArray if dtype is ExtensionDtype, otherwise a NumPy
            ndarray with 'dtype' for its dtype.
        """
        dtype = pd.api.types.pandas_dtype(dtype)
        if pd.core.dtypes.common.is_dtype_equal(dtype, self.dtype):
            if not copy:
                return self
            return self.copy()

        if isinstance(dtype, ExtensionDtype):
            cls = dtype.construct_array_type()
            return cls._from_sequence(self, dtype=dtype, copy=copy)

        return self._data.astype(dtype=dtype, copy=copy)

    def isna(self):
        return np.isnan(self._data)

    def _values_for_argsort(self) -> np.ndarray:
        """Return values for sorting.

        Returns
        -------
        ndarray
            The transformed values should maintain the ordering between values
            within the array.

        See Also
        --------
        ExtensionArray.argsort : Return the indices that would sort this array.

        Notes
        -----
        The caller is responsible for *not* modifying these values in-place, so
        it is safe for implementors to give views on `self`.

        Functions that use this (e.g. ExtensionArray.argsort) should ignore
        entries with missing values in the original array (according to
        `self.isna()`). This means that the corresponding entries in the
        returned array don't need to be modified to sort correctly.
        """
        # Note: this is used in `ExtensionArray.argsort/argmin/argmax`.
        return self._data

    def _values_for_factorize(self) -> tuple:
        """Return an array and missing value suitable for factorization.

        Returns
        -------
        values : ndarray
            An array suitable for factorization. This should maintain order
            and be a supported dtype (Float64, Int64, UInt64, String, Object).
            By default, the extension array is cast to object dtype.
        na_value : object
            The value in `values` to consider missing. This will be treated
            as NA in the factorization routines, so it will be coded as
            `na_sentinel` and not included in `uniques`.

        Notes
        -----
        The values returned by this method are also used in
        :func:`pandas.util.hash_pandas_object`.
        """
        return self._data, self.dtype.na_value

    def take(self, indexer, allow_fill=False, fill_value=None):
        """Take elements from an array.

        Parameters
        ----------
        indices : sequence of int or one-dimensional np.ndarray of int
            Indices to be taken.
        allow_fill : bool, default False
            How to handle negative values in `indices`.
            * False: negative values in `indices` indicate positional indices
              from the right (the default). This is similar to
              :func:`numpy.take`.
            * True: negative values in `indices` indicate
              missing values. These values are set to `fill_value`. Any other
              other negative values raise a ``ValueError``.
        fill_value : any, optional
            Fill value to use for NA-indices when `allow_fill` is True.
            This may be ``None``, in which case the default NA value for
            the type, ``self.dtype.na_value``, is used.
            For many ExtensionArrays, there will be two representations of
            `fill_value`: a user-facing "boxed" scalar, and a low-level
            physical NA value. `fill_value` should be the user-facing version,
            and the implementation should handle translating that to the
            physical version for processing the take if necessary.

        Returns
        -------
        ExtensionArray

        Raises
        ------
        IndexError
            When the indices are out of bounds for the array.
        ValueError
            When `indices` contains negative values other than ``-1``
            and `allow_fill` is True.
    
        See Also
        --------
        numpy.take : Take elements from an array along an axis.
        api.extensions.take : Take elements from an array.

        Notes
        -----
        ExtensionArray.take is called by ``Series.__getitem__``, ``.loc``,
        ``iloc``, when `indices` is a sequence of values. Additionally,
        it's called by :meth:`Series.reindex`, or any other method
        that causes realignment, with a `fill_value`.

        Examples
        --------
        Here's an example implementation, which relies on casting the
        extension array to object dtype. This uses the helper method
        :func:`pandas.api.extensions.take`.

        .. code-block:: python

           def take(self, indices, allow_fill=False, fill_value=None):
               from pandas.core.algorithms import take

               # If the ExtensionArray is backed by an ndarray, then
               # just pass that here instead of coercing to object.
               data = self.astype(object)

               if allow_fill and fill_value is None:
                   fill_value = self.dtype.na_value

               # fill value should always be translated from the scalar
               # type for the array, to the physical storage type for
               # the data, before passing to take.

               result = take(data, indices, fill_value=fill_value,
                             allow_fill=allow_fill)
               return self._from_sequence(result, dtype=self.dtype)
        """
        from pandas.core.algorithms import take

        data = self._data
        if allow_fill and fill_value is None:
            fill_value = self.dtype.na_value

        result = take(
            data,
            indexer,
            fill_value=fill_value,
            allow_fill=allow_fill
        )
        return self._from_sequence(result, dtype=self.dtype)

    def copy(self):
        """Return a copy of the array.

        Returns
        -------
        ExtensionArray
        """
        return type(self)(self._data.copy())

    @classmethod
    def _concat_same_type(cls, to_concat):
        """Concatenate multiple array of this dtype.

        Parameters
        ----------
        to_concat : sequence of this type

        Returns
        -------
        ExtensionArray
        """
        return cls(np.concatenate([x._data for x in to_concat]))

    def _reduce(self, name, skipna=True, **kwargs):
        """Return a scalar result of performing the reduction operation.

        Parameters
        ----------
        name : str
            Name of the function, supported values are:
            { any, all, min, max, sum, mean, median, prod,
            std, var, sem, kurt, skew }.
        skipna : bool, default True
            If True, skip NaN values.
        **kwargs
            Additional keyword arguments passed to the reduction function.
            Currently, `ddof` is the only supported kwarg.

        Returns
        -------
        scalar

        Raises
        ------
        TypeError : subclass does not define reductions
        """
        # NOTE: this implementation is taken from pandas test suite

        if skipna:
            # If we don't have any NAs, we can ignore skipna
            if self.isna().any():
                other = self[~self.isna()]
                return other._reduce(name, **kwargs)

        if name == "sum" and len(self) == 0:
            # GH#29630 avoid returning int 0 or np.bool_(False) on old numpy
            item_type = self._pdcast_type.type_def
            return item_type(0)

        try:
            op = getattr(self.data, name)
        except AttributeError:
            raise NotImplementedError(
                f"{str(self._pdcast_type)} does not support the {name} "
                f"operation"
            )
        return op(axis=0)

    def __array_ufunc__(self, ufunc: np.ufunc, method: str, *inputs, **kwargs):
        """Allows ObjectArrays to utilize numpy ufuncs natively, without
        being coerced to dtype: object.

        This implementation is adapted from the
        :class:`example <numpy.lib.mixins.NDArrayOperatorsMixin>` given by
        numpy.  It might require further modifications.
        """
        # pandas unboxes these so we don't need to implement them ourselves
        if any(
            isinstance(other, (ABCSeries, ABCIndex, ABCDataFrame))
            for other in inputs
        ):
            return NotImplemented

        # get handled values
        array_like = (np.ndarray, ExtensionArray)
        scalar_like = (self._pdcast_type.type_def,)
        if self.dtype._is_numeric:
            scalar_like += (numbers.Number,)

        if not all(isinstance(t, array_like + scalar_like) for t in inputs):
            return NotImplemented

        inputs = tuple(
            x._data if isinstance(x, ExtensionArray) else x for x in inputs
        )
        result = getattr(ufunc, method)(*inputs, **kwargs)

        def reconstruct(x):
            if isinstance(x, scalar_like):
                return x
            return type(self)._from_sequence(x)

        if isinstance(result, tuple):
            return tuple(reconstruct(x) for x in result)
        return reconstruct(result)

    def __arrow_array__(self, type=None):
        """Convert the underlying array values into a pyarrow Array.

        This method ensures that pyarrow knowns how to convert the specific
        extension array into a pyarrow.Array (also when included as a column in
        a pandas DataFrame).
        """
        import pyarrow

        return pyarrow.array(self._data, type=type)

    def __from_arrow__(self, array) -> ExtensionArray:
        """Convert a pyarrow array into a pandas ExtensionArray.

        This method controls the conversion back from pyarrow to a pandas
        ExtensionArray.  It receives a pyarrow Array or ChunkedArray as its
        only argument and is expected to return the appropriate pandas
        ExtensionArray for this dtype and the passed values
        """
        return self._from_sequence(array)

    ###########################
    ####    NEW METHODS    ####
    ###########################

    # NOTE: these are not defined in the base ExtensionArray class, but are
    # called in several pandas operations.

    # def round(self, *args, **kwargs) -> ExtensionArray:
    #     """Default round implementation.  This simply passes through to
    #     np.round().
    #     """
    #     return self._from_sequence(self._data.round(*args, **kwargs))

    # def value_counts(self, dropna: bool = True) -> pd.Series:
    #     """Returns a Series containing counts of each unique value.

    #     Parameters
    #     ----------
    #     dropna : bool, default True
    #         Don't include counts of missing values.

    #     Returns
    #     -------
    #     counts : Series

    #     See Also
    #     --------
    #     Series.value_counts
    #     """
    #     # compute counts without nans
    #     mask = self.isna()
    #     counts = Counter(self._data[~mask])
    #     result = pd.Series(
    #         counts.values(),
    #         index=pd.Index(list(counts.keys()), dtype=self.dtype)
    #     )
    #     if dropna:
    #         return result.astype("Int64")

    #     # if we want to include nans, count mask
    #     nans = pd.Series(
    #         [mask.sum()],
    #         index=pd.Index([self.dtype.na_value], dtype=self.dtype)
    #     )
    #     result = pd.concat([result, nans])
    #     return result.astype("Int64")

    ###############################
    ####    UNARY OPERATORS    ####
    ###############################

    def __neg__(self):
        # NOTE: shim allows unary negation (-)
        if self.dtype._is_boolean:
            return self.__invert__()
        return self._from_sequence(-self._data)

    def __pos__(self):
        # NOTE: shim allows unary +
        return self._from_sequence(+self._data)

    def __invert__(self):
        # NOTE: shim allows unary inversion (~)
        if self.dtype._is_boolean:
            return self.__xor__(True)
        return self._from_sequence(~self._data)

    ################################
    ####    BINARY OPERATORS    ####
    ################################

    def __and__(self, other):
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):
            return NotImplemented

        # NOTE: shim allows binary and (&)
        return self._from_sequence(self._data & other)

    def __xor__(self, other):
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):
            return NotImplemented

        # NOTE: shim allows binary xor (^)
        return self._from_sequence(self._data ^ other)

    def __or__(self, other):
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):
            return NotImplemented

        # NOTE: shim allows binary or (|)
        return self._from_sequence(self._data | other)


# add scalar ops via ExtensionScalarOpsMixin
M8Array._add_arithmetic_ops()
M8Array._add_comparison_ops()



"""Mypy stubs for pdcast/util/round/decimal.pyx"""""
from decimal import Decimal

from pdcast.util.type_hints import array_like

def round_decimal(
    val: Decimal | array_like, rule: str, decimals: int
) -> Decimal | array_like:
    ...




"""This module contains dispatched cast() implementations for floating point
data.
"""
# pylint: disable=unused-argument
import pandas as pd

from pdcast import types
from pdcast.util.numeric import boundscheck
from pdcast.util.round import Tolerance
from pdcast.util.vector import apply_with_errors

from .base import (
    cast, generic_to_boolean, generic_to_integer, snap_round
)


# TODO: float -> float doesn't account for tol


@cast.overload("float", "bool")
def float_to_boolean(
    series: pd.Series,
    dtype: types.ScalarType,
    rounding: str,
    tol: Tolerance,
    errors: str,
    **unused
) -> pd.Series:
    """Convert floating point data to a boolean data type."""
    series = snap_round(
        series,
        tol=tol.real,
        rule=rounding,
        errors=errors
    )
    series = boundscheck(series, dtype=dtype, tol=tol.real, errors=errors)
    return generic_to_boolean(series, dtype, errors=errors)


@cast.overload("float", "int")
def float_to_integer(
    series: pd.Series,
    dtype: types.ScalarType,
    rounding: str,
    tol: Tolerance,
    downcast: types.CompositeType,
    errors: str,
    **unused
) -> pd.Series:
    """Convert floating point data to an integer data type."""
    series = snap_round(
        series,
        tol=tol.real,
        rule=rounding,
        errors=errors
    )
    series = boundscheck(series, dtype=dtype, tol=tol.real, errors=errors)
    return generic_to_integer(
        series,
        dtype,
        tol=tol,
        downcast=downcast,
        errors=errors
    )


try:  # float80 might not be defined on all systems

    @cast.overload("float80", "decimal")
    def longdouble_to_decimal(
        series: pd.Series,
        dtype: types.ScalarType,
        errors: str,
        **unused
    ) -> pd.Series:
        """A special case of float_to_decimal() that bypasses
        `TypeError: conversion from numpy.float128 to Decimal is not supported`.
        """
        # convert longdouble to integer ratio and then to decimal
        def call(element):
            numerator, denominator = element.as_integer_ratio()
            return dtype.type_def(numerator) / denominator

        result = apply_with_errors(series, call, errors=errors)
        return result.astype(dtype.dtype)

except ValueError:
    pass



"""Setup script for Bertrand."""
from pathlib import Path
import os
import subprocess
import sys
from typing import Any

from bertrand import Extension, BuildExt, setup


# NOTE: C++ users have to execute $(python3 -m bertrand -I) to compile against bertrand.h
# g++ foo.cpp -o foo.out $(python3 -m bertrand -I)

# NOTE: Python users writing C++ extensions should use bertrand's custom Extension
# class to add the necessary linkages to the bertrand library, and to enable clangd
# support through compile_commands.json.


# NOTE: Bertrand uses C++23 features only found in GCC 14+ and Clang 18+ (MSVC is not
# yet fully supported).  Users must first install these compilers to use the library.
# Here are the steps to do so:


# pylint: disable=pointless-string-statement

"""
GCC: If you have Ubuntu 24.04 or later, then you can install gcc-14 directly using the
package manager:
    sudo apt-get install build-essential gcc-14 g++-14 python3-dev

Otherwise:

1. set up local build variables:
    root=$(realpath $(pwd))
    gcc_source=${root}/gcc-source
    gcc_build=${root}/gcc-build
    gcc_install=${root}/gcc-install
    cpp_libs=${gcc_install}/lib64
2. clone the git repository to a local directory:
    git clone git://gcc.gnu.org/git/gcc.git $gcc_source
    cd $gcc_source
3. checkout a GCC 14+ release branch:
    git checkout releases/gcc-14
4. download prerequisites
    ./contrib/download_prerequisites
5. create a local build directory and configure the build:
    mkdir $gcc_build
    cd $gcc_build
    ${gcc_source}/configure --prefix=${gcc_install} --enable-languages=c,c++ \
        --enable-shared --disable-werror --disable-multilib --disable-bootstrap
6. run the build (takes ~15 minutes):
    make -j$(nproc)
7. install:
    make install
8. (optional) update alternatives and set as default compiler:
    sudo update-alternatives --install /usr/bin/gcc gcc ${gcc_install}/bin/gcc 14 \
        --slave /usr/bin/g++ g++ ${gcc_install}/bin/g++ \
        --slave /usr/bin/c++ c++ ${gcc_install}/bin/c++ \
        --slave /usr/bin/cpp cpp ${gcc_install}/bin/cpp \
        --slave /usr/bin/gcc-ar gcc-ar ${gcc_install}/bin/gcc-ar \
        --slave /usr/bin/gcc-nm gcc-nm ${gcc_install}/bin/gcc-nm \
        --slave /usr/bin/gcc-ranlib gcc-ranlib ${gcc_install}/bin/gcc-ranlib \
        --slave /usr/bin/gcov gcov ${gcc_install}/bin/gcov
9. Check the version:
    gcc --version

Clang:

1. set up local build variables:
    root=$(realpath $(pwd))
    clang_source=${root}/clang-source
    clang_build=${root}/clang-build
    clang_install=${root}/clang-install
    cpp_libs=${clang_install}/lib64  # TODO: confirm this path
1. clone the git repository to a local directory:
    git clone https://github.com/llvm/llvm-project.git $clang_source
    cd $clang_source
2. checkout the clang 18 release branch:
    git checkout release/18.x
3. create a local build directory and configure the build:
    mkdir $clang_build
    cd $clang_build
    cmake -DLLVM_ENABLE_PROJECTS=clang -DCMAKE_BUILD_TYPE=Release -G "Unix Makefiles" \
        -DCMAKE_INSTALL_PREFIX=${clang_install} -DBUILD_SHARED_LIBS=ON \
        ${clang_source}/llvm
4. run the build (takes ~30 minutes):
    make -j$(nproc)
5. install:
    make install
6. (optional) update alternatives and set as default compiler:
    sudo update-alternatives --install /usr/bin/clang clang ${clang_install}/bin/clang 18 \
        --slave /usr/bin/clang++ clang++ ${clang_install}/bin/clang++ \
        --slave /usr/bin/clang-cpp clang-cpp ${clang_install}/bin/clang-cpp
7. Check the version:
    clang --version
"""


# NOTE: build ninja from source to enable C++20 modules and faster builds.


"""
1. set up local build variables:
    ninja_install=${root}/ninja-install
2. clone the git repository to a local directory:
    git clone https://github.com/ninja-build/ninja.git $ninja_install
    cd $ninja_install
3. checkout the latest release branch:
    git checkout release
4. bootstrap the build:
    ./configure.py --bootstrap
5. (optional) update alternatives and set as default ninja:
    sudo update-alternatives --install /usr/bin/ninja ninja ${ninja_install}/ninja 112
6. Check the version:
    ninja --version
"""


# NOTE: build CMake from source to enable C++20 module support.

"""
1. set up local build variables:
    cmake_source=${root}/cmake-source
    cmake_build=${root}/cmake-build
    cmake_install=${root}/cmake-install
2. clone the git repository to a local directory:
    git clone https://github.com/Kitware/CMake.git $cmake_source
    cd $cmake_source
3. checkout the latest release branch:
    git checkout release
4. create a local build directory and configure the build (takes ~5 minutes):
    mkdir $cmake_build
    cd $cmake_build
    ${cmake_source}/bootstrap --prefix=${cmake_install} --generator="Ninja" \
        LDFLAGS="-Wl,-rpath=${cpp_libs} -L${cpp_libs}"
5. run the build:
    ninja -j$(nproc)
6. install:
    ninja install
7. (optional) update alternatives and set as default cmake:
    sudo update-alternatives --install /usr/bin/cmake cmake ${cmake_install}/bin/cmake 328 \
        --slave /usr/bin/ctest ctest ${cmake_install}/bin/ctest
8. Check the version:
    cmake --version
"""


# NOTE: after building the C++ compiler, it's a good idea to build Python from source
# as well to prevent any ABI incompatibilities and get the best possible performance.
# Here's how to do that:

"""
1. set up local build variables:
    python_source=${root}/python-source
    python_build=${root}/python-build
    python_install=${root}/python-install
2. clone the git repository to a local directory:
    git clone https://github.com/python/cpython.git $python_source
    cd $python_source
3. checkout the latest release branch:
    git checkout 3.12
4. download prerequisites and install dependencies
    sudo apt-get install build-essential gdb lcov pkg-config \
        libbz2-dev libffi-dev libgdbm-dev libgdbm-compat-dev liblzma-dev \
        libncurses5-dev libreadline6-dev libsqlite3-dev libssl-dev \
        lzma lzma-dev tk-dev uuid-dev zlib1g-dev
5. create a local build directory and configure the build:
    mkdir $python_build
    cd $python_build
    ${python_source}/configure --prefix=${python_install} --with-ensurepip=upgrade \
        --enable-shared --enable-optimizations --with-lto \
        LDFLAGS="-Wl,-rpath=${python_install}/lib:${cpp_libs}" \
        CFLAGS="-DBERTRAND_TRACEBACK_EXCLUDE_PYTHON=${python_source}:${python_build}:${python_install}"
6. run the build (takes ~30 minutes):
    make -s -j$(nproc)
7. install:
    make install
8. (optional) update alternatives and set as default Python:
    sudo update-alternatives --install /usr/bin/python python ${python_install}/bin/python3.12 312 \
        --slave /usr/bin/python3 python3 ${python_install}/bin/python3.12 \
        --slave /usr/bin/pip pip ${python_install}/bin/pip3.12 \
        --slave /usr/bin/pip3 pip3 ${python_install}/bin/pip3.12
9. Check the version:
    python --version
"""


# NOTE: You might want to add additional configuration flags to either the C++ compiler
# or the Python build process to enable more warnings, enhanced security, or specific
# optimizations.  Here are some to consider:
#
#   GCC:
#
#   Clang
#
#   Python:
#       --with-valgrind:  enable valgrind support for Python


# Extra:
# -fno-strict-overflow

# Missing:
# -fwrapv
# -fstack-protector-strong
# -Wformat
# -Werror=format-security
# -Wdate-time
# -D_FORTIFY_SOURCE=2



# NOTE: See setuptools for more info on how extensions are built:
# https://setuptools.pypa.io/en/latest/userguide/ext_modules.html


class build_ext(BuildExt):
    """A custom build_ext command that installs third-party C++ packages and builds C++
    unit tests as part of pip install.
    """

    def extract(self, cwd: Path, tarball: str) -> None:
        """Extract a tarball in the current working directory.

        Parameters
        ----------
        cwd : Path
            A path to the directory where the tarball is located.
        tarball : str
            The name of the tarball to extract.
        """
        try:
            subprocess.check_call(["tar", "-xzf", tarball], cwd=str(cwd))
        except subprocess.CalledProcessError as exc:
            print(f"failed to extract {tarball}:", exc)
            sys.exit(1)

    def install_cpptrace(self, cwd: Path) -> None:
        """Build cpptrace from source after extracting it.

        Parameters
        ----------
        cwd : Path
            A path to the directory where the tarball was extracted.
        """
        cpptrace_dir = "cpptrace-0.5.2"
        if (cwd / cpptrace_dir).exists():
            cwd = cwd / cpptrace_dir
        else:
            self.extract(cwd, f"{cpptrace_dir}.tar.gz")
            subprocess.check_call(["mv", "cpptrace", cpptrace_dir], cwd=str(cwd))
            cwd = cwd / cpptrace_dir / "build"
            cwd.mkdir(parents=True, exist_ok=True)

            try:
                subprocess.check_call(
                    ["cmake", "..", "-DCMAKE_BUILD_TYPE=Release"],
                    cwd=str(cwd)
                )
                subprocess.check_call(["make"], cwd=str(cwd))
            except subprocess.CalledProcessError as exc:
                print("failed to build cpptrace:", exc)
                sys.exit(1)

            cwd = cwd.parent  # back out to root dir

        # add headers to include path and link against binary
        self.include_dirs.append(str(cwd / "include"))
        self.library_dirs.append(str(cwd / "build"))
        self.library_dirs.append(str(cwd / "build/_deps/libdwarf-build/src/lib/libdwarf"))
        self.library_dirs.append(str(cwd / "build/_deps/zstd-build/lib"))
        self.libraries.extend(["cpptrace", "dwarf", "z", "zstd", "dl"])

    def install_gtest(self, cwd: Path) -> None:
        """Build GoogleTest from source after extracting it.

        Parameters
        ----------
        cwd : Path
            A path to the directory where the tarball was extracted.
        """
        gtest_dir = "googletest-1.14.0"
        if (cwd / gtest_dir).exists():
            cwd = cwd / gtest_dir
        else:
            self.extract(cwd, f"{gtest_dir}.tar.gz")
            cwd = cwd / gtest_dir / "build"
            cwd.mkdir(parents=True, exist_ok=True)

            try:
                subprocess.check_call(
                    ["cmake", "..", "-DCMAKE_CXX_FLAGS=-fPIC"],
                    cwd=str(cwd)
                )
                subprocess.check_call(["make"], cwd=str(cwd))
            except subprocess.CalledProcessError as exc:
                print("failed to build GoogleTest:", exc)
                sys.exit(1)

            cwd = cwd.parent  # back out to root dir

        # add headers to include path and link against binary
        self.include_dirs.append(str(cwd / "googletest/include"))
        self.library_dirs.append(str(cwd / "build/lib"))
        self.libraries.append("gtest")

    def install_pcre2(self, cwd: Path) -> None:
        """Build PCRE2 from source after extracting it.

        Parameters
        ----------
        cwd : Path
            A path to the directory where the tarball was extracted.
        """
        pcre_dir = "pcre2-10.43"
        if (cwd / pcre_dir).exists():
            cwd = cwd / pcre_dir
        else:
            self.extract(cwd, f"{pcre_dir}.tar.gz")
            cwd = cwd / pcre_dir
            cwd.mkdir(parents=True, exist_ok=True)

            try:
                subprocess.check_call(["./configure", "--enable-jit=auto"], cwd=str(cwd))
                subprocess.check_call(["make"], cwd=str(cwd))
            except subprocess.CalledProcessError as exc:
                print("failed to build PCRE2:", exc)
                sys.exit(1)

        # add headers to include path and link against binary
        self.include_dirs.append(str(cwd / "src"))
        self.library_dirs.append(str(cwd / ".libs"))
        self.libraries.append("pcre2-8")

    def run(self, *args: Any, **kwargs: Any) -> None:
        """Build third-party libraries from source before installing any extensions.

        Parameters
        ----------
        *args : Any
            Arbitrary positional arguments passed to the superclass.
        **kwargs : Any
            Arbitrary keyword arguments passed to the superclass.
        """
        try: # check for cmake
            subprocess.check_call(
                ["cmake", "--version"],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL
            )
        except subprocess.CalledProcessError:
            print("CMake not installed")
            sys.exit(1)

        # install third-party C++ packages
        cwd = Path(os.getcwd()) / "third_party"
        self.install_cpptrace(cwd)
        self.install_pcre2(cwd)
        self.install_gtest(cwd)

        # compile Python extensions
        super().run(*args, **kwargs)

        # save compiler flags to test/ so that we can use the same configuration when
        # building the test suite
        cwd = cwd.parent / "test"
        with (cwd / ".compile_flags").open("w") as file:
            file.write(" ".join(
                self.compiler.compiler_so[1:] +  # remove the compiler name
                ["-std=c++20", "-fvisibility=hidden", "-g0"]  # added by Pybind11Extension
            ))


EXTENSIONS = [
    # Pybind11Extension(
    #     "bertrand.structs.linked",
    #     sources=["bertrand/structs/linked.cpp"],
    #     extra_compile_args=["-O3"]
    # ),
    # Extension(
    #     "*",
    #     sources=["bertrand/**/*.pyx"],
    #     define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
    # ),
    Extension("example", ["example.cpp", "example_module.cpp"], executable=True),
    # Extension("example_exe", ["example_exe.cpp"], executable=True),
    # Extension("bertrand.regex", ["bertrand/regex.cpp"])
]


setup(
    # ext_modules=EXTENSIONS,
    # cmdclass={"build_ext": build_ext},
)



"""This module contains the base class for all bertrand scalar types.

The entire contents of this file are hosted in the online documentation, and serve as
a reference for defining and registering new types within the bertrand type system.
Users can refer to this file to see how to build new types.
"""
from __future__ import annotations

from collections import deque
from typing import Any

import numpy as np
import pandas as pd

from .meta import DTYPE, EMPTY, POINTER_SIZE, Base, Empty, TypeMeta


# TODO: probably just put this back in meta.py and then provide a simple example here


# pylint: disable=unused-argument


def _check_scalar(
    value: Any | Empty,
    namespace: dict[str, Any],
    processed: dict[str, Any],
) -> type:
    """Validate a scalar Python type provided in a bertrand type's namespace or infer
    it from a provided dtype.
    """
    if "scalar" in processed:  # auto-generated in check_dtype
        return processed["scalar"]

    dtype = namespace.get("dtype", EMPTY)

    if value is EMPTY:
        if dtype is EMPTY:
            raise TypeError("type must define at least one of 'dtype' and/or 'scalar'")
        if not isinstance(dtype, (np.dtype, pd.api.extensions.ExtensionDtype)):
            raise TypeError(f"dtype must be a numpy/pandas dtype, not {repr(dtype)}")

        processed["dtype"] = dtype
        return dtype.type

    if not isinstance(value, type):
        raise TypeError(f"scalar must be a Python type object, not {repr(value)}")

    if dtype is EMPTY:
        processed["dtype"] = None  # TODO: synthesize dtype
    else:
        processed["dtype"] = dtype
    # elif value == dtype.type:
    #     processed["dtype"] = dtype
    # else:
    #     raise TypeError(
    #         f"scalar must be consistent with dtype.type: {repr(value)} != "
    #         f"{repr(dtype.type)}"
    #     )

    return value


def _check_dtype(
    value: Any | Empty,
    namespace: dict[str, Any],
    processed: dict[str, Any]
) -> DTYPE:
    """Validate a numpy or pandas dtype provided in a bertrand type's namespace or
    infer it from a provided scalar.
    """
    if "dtype" in processed:  # auto-generated in check_scalar
        return processed["dtype"]

    scalar = namespace.get("scalar", EMPTY)

    if value is EMPTY:
        if scalar is EMPTY:
            raise TypeError("type must define at least one of 'dtype' and/or 'scalar'")
        if not isinstance(scalar, type):
            raise TypeError(f"scalar must be a Python type object, not {repr(scalar)}")

        processed["scalar"] = scalar
        return None  # NOTE: causes dtype to be synthesized during builder.register()

    if not isinstance(value, (np.dtype, pd.api.extensions.ExtensionDtype)):
        raise TypeError(f"dtype must be a numpy/pandas dtype, not {repr(value)}")

    if scalar is EMPTY:
        processed["scalar"] = value.type
    else:
        processed["scalar"] = scalar
    # elif value.type == scalar:
    #     processed["scalar"] = scalar
    # else:
    #     raise TypeError(
    #         f"dtype.type must be consistent with scalar: {repr(value.type)} != "
    #         f"{repr(scalar)}"
    #     )

    return value


def _check_itemsize(
    value: Any | Empty,
    namespace: dict[str, Any],
    processed: dict[str, Any]
) -> int:
    """Validate an itemsize provided in a bertrand type's namespace."""
    if value is EMPTY:
        if "dtype" in namespace:
            dtype = namespace["dtype"]
            if dtype is None:
                return POINTER_SIZE
            return dtype.itemsize

        if "dtype" in processed:
            dtype = processed["dtype"]
            if dtype is None:
                return POINTER_SIZE
            return dtype.itemsize

    if not isinstance(value, int) or value <= 0:
        raise TypeError(f"itemsize must be a positive integer, not {repr(value)}")

    return value


def _check_max(
    value: Any | Empty,
    namespace: dict[str, Any],
    processed: dict[str, Any]
) -> int | float:
    """Validate a maximum value provided in a bertrand type's namespace."""
    if "max" in processed:  # auto-generated in check_min
        return processed["max"]

    min_val = namespace.get("min", -np.inf)

    if value is EMPTY:
        processed["min"] = min_val
        return np.inf

    if not (isinstance(value, int) or isinstance(value, float) and value == np.inf):
        raise TypeError(f"min must be an integer or infinity, not {repr(value)}")

    if value < min_val:
        raise TypeError(f"max must be greater than min: {value} < {min_val}")

    processed["min"] = min_val
    return value


def _check_min(
    value: Any | Empty,
    namespace: dict[str, Any],
    processed: dict[str, Any]
) -> int | float:
    """Validate a minimum value provided in a bertrand type's namespace."""
    if "min" in processed:  # auto-generated in check_max
        return processed["min"]

    max_val = namespace.get("max", np.inf)

    if value is EMPTY:
        processed["max"] = max_val
        return -np.inf

    if not (isinstance(value, int) or isinstance(value, float) and value == -np.inf):
        raise TypeError(f"min must be an integer or infinity, not {repr(value)}")

    if value > max_val:
        raise TypeError(f"min must be less than or equal to max: {value} > {max_val}")

    processed["max"] = max_val
    return value


def _check_is_nullable(
    value: Any | Empty,
    namespace: dict[str, Any],
    processed: dict[str, Any]
) -> bool:
    """Validate a nullability flag provided in a bertrand type's namespace."""
    return True if value is EMPTY else bool(value)


def _check_is_numeric(
    value: Any | Empty,
    namespace: dict[str, Any],
    processed: dict[str, Any]
) -> bool:
    """Validate a numeric flag provided in a bertrand type's namespace."""
    return False if value is EMPTY else bool(value)


def _check_missing(
    value: Any | Empty,
    namespace: dict[str, Any],
    processed: dict[str, Any]
) -> Any:
    """Validate a missing value provided in a bertrand type's namespace."""
    if value is EMPTY:
        return pd.NA

    if not pd.isna(value):  # type: ignore
        raise TypeError(f"missing value must pass a pandas.isna() check: {repr(value)}")

    return value


class Type(Base, metaclass=TypeMeta):
    """Parent class for all scalar types.

    Inheriting from this class triggers the metaclass machinery, which automatically
    validates and registers the inheriting type.  As such, any class that inherits from
    this type becomes the root of a new type hierarchy, which can be extended to create
    tree structures of arbitrary depth.
    """

    # NOTE: every type has a collection of strings, python types, and numpy/pandas
    # dtype objects which are used to identify the type during detect() and resolve()
    # calls.  Here's how they work
    #   1.  Python types (e.g. int, str, etc.) - used during elementwise detect() to
    #       identify scalar values.  If defined, the type's `from_scalar()` method will
    #       be called automatically to parse each element.
    #   2.  numpy/pandas dtype objects (e.g. np.dtype("i4")) - used during detect() to
    #       identify array-like containers that implement a `.dtype` attribute.  If
    #       defined, the type's `from_dtype()` method will be called automatically to
    #       parse the dtype.  The same process also occurs whenever resolve() is called
    #       with a literal dtype argument.
    #   3.  Strings - used during resolve() to identify types by their aliases.
    #       Optional arguments can be provided in square brackets after the alias, and
    #       will be tokenized and passed to the type's `from_string()` method, if it
    #       exists.  Note that a type's aliases implicitly include its class name, so
    #       that type hints containing the type can be identified when
    #       `from __future__ import annotations` is enabled.  It is good practice to
    #       follow suit with other aliases so they can be used in type hints as well.

    aliases = {"foo", int, type(np.dtype("i4"))}  # <- for example (not evaluated)

    # NOTE: Any field assigned to EMPTY is a required attribute that must be defined
    # by concretions of this type.  Each field can be inherited from a parent type
    # with normal semantics.  If a type does not define a required attribute, then the
    # validation function will receive EMPTY as its value, and must either raise an
    # error or replace it with a default value.

    scalar:         type            = EMPTY(_check_scalar)  # type: ignore
    dtype:          DTYPE           = EMPTY(_check_dtype)  # type: ignore
    itemsize:       int             = EMPTY(_check_itemsize)  # type: ignore
    max:            int | float     = EMPTY(_check_max)  # type: ignore
    min:            int | float     = EMPTY(_check_min)  # type: ignore
    is_nullable:    bool            = EMPTY(_check_is_nullable)  # type: ignore
    is_numeric:     bool            = EMPTY(_check_is_numeric)  # type: ignore
    missing:        Any             = EMPTY(_check_missing)

    # NOTE: the following methods can be used to allow parametrization of a type using
    # an integrated flyweight cache.  At minimum, a parametrized type must implement
    # `__class_getitem__`, which specifies the parameter names and default values, as
    # well as `from_string`, which parses the equivalent syntax in the
    # type-specification mini-language.  The other methods are optional, and can be
    # used to customize the behavior of the detect() and resolve() helper functions.

    # None of these methods will be inherited by subclasses, so their behavior is
    # unique for each type.  If they are not defined, they will default to the
    # identity function, which will be optimized away where possible to improve
    # performance.

    def __class_getitem__(cls, spam: str = "foo", eggs: int = 2) -> TypeMeta:
        """Example implementation showing how to create a parametrized type.

        When this method is defined, the metaclass will analyze its signature and
        extract any default values, which will be forwarded to the class's
        :attr:`params <Type.params>` attribute, as well as its base namespace.  Each
        argument must have a default value, and the signature must not contain any
        keyword or variadic arguments (i.e. *args or **kwargs).

        Positional arguments to the flyweight() helper method should be supplied in
        the same order as they appear in the signature, and will be used to form the
        string identifier for the type.  Keywords are piped directly into the resulting
        class's namespace, and will override any existing values.  In the interest of
        speed, no checks are performed on the arguments, so it is up to the user to
        ensure that they are in the expected format.
        """
        return cls.flyweight(spam, eggs, dtype=np.dtype("i4"))

    @classmethod
    def from_string(cls, spam: str, eggs: str) -> TypeMeta:
        """Example implementation showing how to parse a string in the
        type-specification mini-language.

        This method will be called automatically whenever :func:`resolve` encounters
        an alias that matches this type.  Its signature must match that of
        :meth:`__class_getitem__`, and it will be invoked with the comma-separated
        tokens parsed from the alias's argument list.  They are always provided as
        strings, with no further processing other than stripping leading/trailing
        whitespace.  It is up to the user to parse them into the appropriate type.
        """
        return cls.flyweight(spam, int(eggs), dtype=np.dtype("i4"))

    @classmethod
    def from_scalar(cls, scalar: Any) -> TypeMeta:
        """Example implementation showing how to parse a scalar value into a
        parametrized type.

        This method will be called during the main :func:`detect` loop whenever a
        scalar or unlabeled sequence of scalars is encountered.  It should be fast,
        since it will be called for every element in the array.  If left blank, it will
        """
        return cls.flyweight(scalar.spam, scalar.eggs(1, 2, 3), dtype=scalar.ham)

    @classmethod
    def from_dtype(cls, dtype: Any) -> TypeMeta:
        """Example implementation showing how to parse a numpy/pandas dtype into a
        parametrized type.
        
        This method will be called whenever :func:`detect` encounters array-like data
        with a related dtype, or when :func:`resolve` is called with a dtype argument.
        """
        return cls.flyweight(dtype.unit, dtype.step_size, dtype=dtype)

    # NOTE: the following methods are provided as default implementations inherited by
    # descendent types.  They can be overridden to customize their behavior, although
    # doing so is not recommended unless absolutely necessary.  Typically, users should
    # invoke the parent method via super() and modify the input/output, rather than
    # reimplementing the method from scratch.

    @classmethod
    def replace(cls, *args: Any, **kwargs: Any) -> TypeMeta:
        """Base implementation of the replace() method, which is inherited by all
        descendant types.
        """
        forwarded = dict(cls.params)
        positional = deque(args)
        n  = len(args)
        i = 0
        for k in forwarded:
            if i < n:
                forwarded[k] = positional.popleft()
            elif k in kwargs:
                forwarded[k] = kwargs.pop(k)

        if positional:
            raise TypeError(
                f"replace() takes at most {len(cls.params)} positional arguments but "
                f"{n} were given"
            )
        if kwargs:
            singular = len(kwargs) == 1
            raise TypeError(
                f"replace() got {'an ' if singular else ''}unexpected keyword argument"
                f"{'' if singular else 's'} [{', '.join(repr(k) for k in kwargs)}]"
            )

        return cls.base_type[*forwarded.values()]



"""This module contains general vector manipulations for 1D numpy arrays and
pandas Series objects.
"""
cimport cython
cimport numpy as np
import numpy as np
import pandas as pd


######################
####    PUBLIC    ####
######################


cpdef np.ndarray as_array(object data):
    """Convert an arbitrary object into a corresponding 1D numpy array.

    Scalars and non-numpy/pandas iterables are converted into dtype: object
    array.
    """
    if isinstance(data, pd.Series):
        return data.to_numpy()

    if isinstance(data, np.ndarray):
        return np.atleast_1d(data)

    return np.atleast_1d(np.array(data, dtype=object))


cpdef object as_series(object data):
    """Convert an arbitrary object into a corresponding pandas Series.

    Scalars and non-numpy/pandas iterables are converted into dtype: object
    series.
    """
    if isinstance(data, pd.Series):
        return data

    if isinstance(data, np.ndarray):
        return pd.Series(np.atleast_1d(data))

    return pd.Series(data, dtype=object)


cpdef object apply_with_errors(object series, object call, str errors):
    """Apply a callable over a series using the specified error handling rule.

    This always returns a `dtype: object` series.

    Parameters
    ----------
    series : pd.Series
        A Series to iterate over.
    call : Callable
        A callable to apply at each index.
    errors : str
        The rule to use if errors are encountered.  Must be one of "raise",
        "ignore", or "coerce".  Both "raise" and "ignore" propagate errors up
        the stack, while "coerce" drops the offending values from the output.

    Returns
    -------
    pd.Series
        The result of applying ``call`` at each index.
    """
    cdef np.ndarray[object] result
    cdef bint has_errors
    cdef np.ndarray[np.uint8_t, cast=True] error_index
    cdef object series_index

    # loop over numpy array
    result, has_errors, error_index = _apply_with_errors(
        series.to_numpy(dtype=object),
        call=call,
        errors=errors
    )

    # remove coerced NAs.  NOTE: has_errors=True only when errors="coerce"
    series_index = series.index
    if has_errors:
        result = result[~error_index]
        series_index = series_index[~error_index]

    return pd.Series(
        result,
        index=series_index,
        name=series.name,
        dtype=object
    )


#######################
####    PRIVATE    ####
#######################


@cython.boundscheck(False)
@cython.wraparound(False)
cdef tuple _apply_with_errors(
    np.ndarray[object] arr,
    object call,
    str errors
):
    """Helper for the apply step of `apply_with_errors()`."""
    cdef unsigned int arr_length = arr.shape[0]
    cdef unsigned int i
    cdef np.ndarray[object] result = np.full(arr_length, None, dtype=object)
    cdef bint has_errors = False
    cdef np.ndarray[np.uint8_t, cast=True] index

    # index is only necessary if errors="coerce"
    if errors == "coerce":
        index = np.full(arr_length, False)
    else:
        index = None

    # apply `call` at every index of array and record errors
    for i in range(arr_length):
        try:
            result[i] = call(arr[i])
        except (KeyboardInterrupt, MemoryError, SystemError, SystemExit):
            raise  # never coerce on these error types
        except Exception as err:
            if errors == "coerce":
                has_errors = True
                index[i] = True
                continue
            raise err

    return result, has_errors, index
